//*******************************************************
//** ISAM Class - Base Class Interface to VSAM Lib
//*-----------------------------------------------------------
//* Written By:  Jay J Falconer
//*              Bitwise Software International, Inc.
//*              Phoenix, AZ 85028
//*              www.ShoppingQ.com
//*
//*              V1.1 - Release Date 8/24/04
//*-----------------------------------------------------------
//* TRACER NOTES:
//*-----------------------------------------------------------
//* The Trace() functions here provide the ability to debug
//* the code at runtime. Step logging will be done if the
//* class code finds the file named in DEBUG_FILE_NAME 
//* in the same directory as the compiled script program. 
//* If this file is missing, the code will not log its steps and calls. 
//* Since it's very difficult to perform debugging in a stateless web
//* environment when a web called program only runs for a 
//* moment, this feature is important to help diagnose
//* problems during end-user runtime. Logging will be 
//* done to a text file called "logfile.txt". Since
//* the logging will slow down the code, it's recommended
//* that the debug,txt file NOT exist unless you need
//* to locate a problem during development or runtime. 
//* While the extra Tracer code does make the object code
//* much bigger, the tradeoff in being able to remotely
//* debug the usage of this code at runtime is well worth it. 
//*-----------------------------------------------------------
//* CODING STYLE:
//*-----------------------------------------------------------
//* While there are many ways to code the same task and there
//* are certainly MUCH shorter ways to accomplish the tasks below, 
//* The coding methods used here are to make the code 
//* as readable and maintainable as possible. It is presented
//* as both a useable production tool as well as a VSAM C++ 
//* implementation tutotial. I also use this code in my beginning
//* C++ classes to show my students the basics of building C++ 
//* classes as well as the power and usability of the STRING class 
//* in general. Of course, we use VSAM for all our embedded DB work. 
//*-----------------------------------------------------------
//* NO WARRANTY:
//*-----------------------------------------------------------
//* This code is provided as part of VSAM for Linux
//* It is provided "AS IS" and the Author does not offer support 
//* nor any warranties or guarantees. 
//* USE THIS CODE at your own risk. 
//*******************************************************
#include "ISAM.h"

//*******************************************************
//* ISAM INIT: 
//*------------------------------------------------------
//*	  MUST be called FIRST in Constructor of Parent Class
//*	  to setup & initialize the file specs and vectors
//*******************************************************
bool ISAM::ClassInit (string file_path_name, string file_name_str, unsigned int max_value = 100)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::ClassInit for file: " + file_path_name + file_name_str);
		Trace ("--> max_value: " + int_to_string (max_value));
	#endif

	reset_error_codes();

	PATH_NAME = file_path_name;		// example Qfiles/ (TRAILING SLASH needed)
	FILE_NAME = file_name_str;		// example Quser (no file suffix)

	if (FILE_NAME.length() == 0)
	{
		set_error (FILE_NAME_INVALID, "--> File Name Invalid or Missing");
		return (false);
	}

	// initialize arrays to be used
	MAX_NUM_FIELDS = max_value;

	// log the max limits imposed by the vector class on this platform
	#ifdef TRACER_CODE
		Trace ("--> max_size for STRING vector: " + long_to_string (F_DATA.max_size()));
		Trace ("--> max_size for INT vector: "    + long_to_string (F_LENGTH.max_size()));

		// more efficient to resize vector sizes here using known max num fields than to keep 
		// using vector.push_back() functions
		Trace ("--> dynamically allocate vectors for DB fields");
	#endif

	F_DATA.resize    (MAX_NUM_FIELDS + 1);	// type = string
	F_NAME.resize    (MAX_NUM_FIELDS + 1);	// type = string
	F_TYPE.resize    (MAX_NUM_FIELDS + 1);	// type = string
	F_INDEX.resize   (MAX_NUM_FIELDS + 1);	// type = string
	F_PARTS.resize   (MAX_NUM_FIELDS + 1);	// type = string
	F_LENGTH.resize  (MAX_NUM_FIELDS + 1);	// type = int
	F_IN_C_KEY.resize(MAX_NUM_FIELDS + 1);	// type = bool

	// INIT of string vectors not needed since strings have their own constructors (= "")
	// But we do need to init all other vectors of non-class types (e.g. built in types)
	#ifdef TRACER_CODE
		Trace ("--> init non-str vectors");
	#endif

	for (unsigned int i = 1; i <= MAX_NUM_FIELDS; i++)
	{
		//		F_DATA       [i] = "";	// data field string content
		//		F_NAME       [i] = "";	// names for each field
		//		F_TYPE       [i] = "";	// type of field S, I, L, F
		//		F_INDEX      [i] = "";	// Y=yes, N=no, P=primary key, "C"=compund key
		//		F_PARTS      [i] = "";	// for compound keys, holds field numbers for VSAM format string 

		F_LENGTH   [i] = 0;			// max length
		F_IN_C_KEY [i] = false;		// part of a compound_key
	}

	#ifdef TRACER_CODE
		Trace ("--> ISAM::init DONE." + trace_file_name ());
	#endif


	return (true);
}

//*******************************************************
//* ISAM Class Constructor
//*******************************************************
ISAM::ISAM (void)
{
// for runtime code debug/Trace functions
	TRACER_MODE         = false;		// must be first thing done if we want the
	TRACER_MODE_CHECKED = false;		// runtime Tracer functions to work...
	TRACER_OFF          = false;

	TRACER_LOGFILE_NAME = "logfile.txt";	// where the step logging goes
	DEBUG_FILE_NAME     = "debug.txt";		// if exists, step logging is done (when tracer code compiled herein)

// end embedded Tracer functions
	#ifdef TRACER_CODE
		Trace ("ISAM class constructor called.");
		Trace ("--> Initialize ISAM Class Variables...");
	#endif

	FILE_HANDLE = 0;
	FILE_OPEN   = false;			// T or F

	VSAM_BUFFER = NULL;			// clear far * to structure of type GSTR

	VSAM_LOCK_KEY = "";			// clear at start

	rec_loaded = false;		 	// T or F

	// the LIC Key provided by Mr. Tony at SoftwareSource
	VSAM_PASSWORD   = "xxxxxxxxxxxxxxxxxxxx";	//  INSERT YOUR KEY HERE

	// Your Own Personal Encrpytion Key for Max Security
	// Can be any string of Characters/Numbers - make it at least 50 characters long
	ENCRYPTION_KEY  = "019298ejiwje009i4uti09fjwdasdvlnaeriovmnqrv3ejg3q0ergj9j09ajerg12we34rt56";

	MAX_STRING_LEN = sizeof(STRING_BUFFER) - 1;

	#ifdef TRACER_CODE
		Trace ("--> NOTE: MAX_STRING_LEN set to: " + int_to_string (MAX_STRING_LEN));
	#endif

	NUM_FIELDS_LOADED = 0;					// excluding primary key field

	error   = false;
	err_num = 0;
	err_msg = "";

	P_KEY_FIELD_LEN   = 0;
	P_KEY_FIELD_NAME  = "";
	P_KEY_FIELD_VALUE = "";

	PRIMARY_KEY = "PRIMARY";			// load this string value for class global use

	CURR_INDEX_NAME = PRIMARY_KEY;		// default is the primary index "PRIMARY"
	CURR_INDEX_NUM  = 0;

//	F_FORMAT = "";				// field format data string
	#ifdef TRACER_CODE
		Trace ("--> ISAM class constructor DONE.");
	#endif
}

//*******************************************************
//* ISAM Class Destructor
//*******************************************************
ISAM::~ISAM (void)
{
	#ifdef TRACER_CODE
		Trace ("ISAM Class DESTRUCTOR Called" + trace_file_name ());
	#endif

	unlock (NO_LOG_ERROR);		// free any record locks that may still exist

	FreeVSAMBuffer();				// free VSAM RECORD BUFFER memory

// Vector class cleanup is compiler automatic (DESTRUCTOR) when this class goes out of scope
// so no need to manually cleanup here as would be the case if we used true dynamic arrays. 
// Better to use vectors since cleanup is automatic. Insures no memory leaks. However, this
// does make the object code bigger.

	close (NO_LOG_ERROR);		// close datafile - if open for some reason

	#ifdef TRACER_CODE
		Trace ("ISAM Class DESTRUCTOR DONE." + trace_file_name ());
	#endif
}

//*******************************************************
//* Initializes the File Format Specs 
//*******************************************************
//* Creates the file if missing on disk.
//* Checks each field spec defined and adds it to the
//* VSAM DB structure if not already defined.  
//* Will also look for and change a field's name if need be
//*******************************************************
bool ISAM::FileConfig (bool build_file_if_missing)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::FileConfig called" + trace_file_name ());
		Trace ("--> MAX_NUM_FIELDS: "    + int_to_string (MAX_NUM_FIELDS));
		Trace ("--> NUM_FIELDS_LOADED: " + int_to_string (NUM_FIELDS_LOADED));
	#endif


	reset_error_codes();

	string field_name;
	string field_type;
	bool   field_is_index;
	long   field_width;
	long   field_dimension;
	string field_justify;

	long field_num_assigned;

// VSAM EXTREME NO LONGER USES A FILE FORMAT STRING when
// creating a new file. Instead, we APPEND each new field 
// definition with VsamAddField call
	if (build_file_if_missing == true)
	{
		if (does_isam_file_exist (PATH_NAME, FILE_NAME) == false)
		{
			create_isam_file ();
		}
	}

// MAKE SURE FILE IS OPEN BEFORE WE CALL any VSAM_AddFieldDef below!!!
// This loop checks each field to see if new fields need to be 
// added to the VSAM rec structure or if any field defs (names/width/index flag) have changed
	if (FileOpen () == true)
	{
		// define variables needed in the loop below
		string Fname;
		string Ftype;
		string Findex;
		string Fnumber;
		string Fclass;
		string Fwidth;
		string Fjust;
		string Fdelete;

		for (unsigned int i = 1; i <= NUM_FIELDS_LOADED; i++)
		{
			field_is_index     = false;	
			field_dimension    = 0;
			field_justify      = "L";			// default is left justify
			field_num_assigned = 0;
			field_width        = 0;

			// prepare variables for possible VsamAddField call
			field_name = F_NAME [i];

			field_type = get_vsam_field_type (F_TYPE [i]);

			if (F_INDEX [i] == "Y")			// regular index field
				field_is_index = true;

			if (F_INDEX [i] == "C")			// add COMPOUND INDEX type --> NOTE - index fields MUST be string types
			{
				field_is_index = true;
				field_type = "C" + F_PARTS [i];
			}

			// load field length for keys and fixed length strings
			field_width = F_LENGTH [i];

			// --------------------------------------------------------------------------------------------
			// See if this field number exists in the data structure or 
			// has had its name changed - THIS MUST BE DONE BY FIELD NUMBER
			// for the logic here to work correctly. 
			//
			// DO NOT use the VsamDelete function to delete a field definition or else the class's 
			// field order assumptions below will not work!
			//
			// NOTE: that primary key field is always field # zero and the rest of the fields are 1...nnn
			// --------------------------------------------------------------------------------------------

			Fnumber = "%"+ int_to_string (i);			

			Fname = GetFieldAttribute (Fnumber, "Fnam");

			bool result = false;

			// check for NEW FIELD detected (better be at end of field list!) 
			if (Fname.length() == 0)	
			{
				#ifdef TRACER_CODE
					Trace ("--> NEW FIELD detected for field_num: " + Fnumber + " (" + field_name + ")");
				#endif

				// perform VSAM call to add new field def
				result = VSAM_AddFieldDef (field_name, field_type, field_is_index, field_dimension, field_width, field_justify, field_num_assigned);

				#ifdef TRACER_CODE
					if (result == true)
						Trace ("--> VSAM_AddFieldDef call successful for " + field_name);
					else
						Trace ("--> VSAM_AddFieldDef call FAILED for " + field_name);
				#endif
			}
			else						// else field number alread exists, has its NAME changed?
			{
				#ifdef TRACER_CODE
					Trace ("--> Field exists for field_num: " + Fnumber + " (" + field_name + ")");
				#endif

				if (strings_dont_match (field_name, Fname))
				{
					#ifdef TRACER_CODE
						Trace ("--> Field name change needed for field_num: " + Fnumber + " (" + Fname + ")");
					#endif

					if (VSAM_ChangeFieldAttribute (Fnumber, "Fnam", field_name) == true)
						Fname = field_name;
				}

				// now get the rest of the attributes by Fnumber (DO NOT fetch Fnumber here!)
				Fclass  = GetFieldAttribute (Fnumber, "Fcls");	// D = data
				Ftype   = GetFieldAttribute (Fnumber, "Ftyp");	// $ = type
				Findex  = GetFieldAttribute (Fnumber, "Find");	// Index "T" or "F"
				Fwidth  = GetFieldAttribute (Fnumber, "Fwid");	// 0 (no limit) or a number 01-xxx
				Fjust   = GetFieldAttribute (Fnumber, "Fjst");	// "L", "C", or "R"
				Fdelete = GetFieldAttribute (Fnumber, "Fdel");	// "T" or "F"

				// Has the index flag changed for EXISTING field?
				if (field_is_index == true && strings_dont_match (Findex, "T") && strings_dont_match (Findex, "P"))
				{
					#ifdef TRACER_CODE
						Trace ("--> Field Index Flag Change to PARTIAL index for field_num: " + Fnumber + " (" + field_name + ")");
					#endif

					if (VSAM_ChangeFieldAttribute (Fnumber, "Find", "P") == true)
						Findex = "P";
				}
				if (field_is_index == false && strings_dont_match (Findex, "F") && strings_dont_match (Findex, "D"))
				{
					#ifdef TRACER_CODE
						Trace ("--> Field Index Flag Change to DISABLED index for field_num: " + Fnumber + " (" + field_name + ")");
					#endif

					if (VSAM_ChangeFieldAttribute (Fnumber, "Find", "D") == true)
						Findex = "D";
				}
				
				// Has the width (field or index length) changed?
				if (field_width != atol(Fwidth.c_str()))
				{
					#ifdef TRACER_CODE
						Trace ("--> Field width change for field_num: " + Fnumber + " (" + field_name + ")");
						Trace ("--> Class Width: " + long_to_string (field_width) + ", VSAM Width: " + Fwidth);
					#endif

					if (VSAM_ChangeFieldAttribute (Fnumber, "Fwid", long_to_string(field_width)) == true)
						Fwidth = long_to_string(field_width);
				}
				
				#ifdef TRACER_CODE
					Trace ("============================================================");
					Trace ("Current Field Specs for " + Fnumber + trace_file_name());
					Trace ("============================================================");
					Trace ("Fname:   " + Fname);
					Trace ("Fnumber: " + Fnumber);
					Trace ("Fclass:  " + Fclass);
					Trace ("Ftype:   " + Ftype);
					Trace ("Findex:  " + Findex);
					Trace ("Fwidth:  " + Fwidth);
					Trace ("Fjust:   " + Fjust);
					Trace ("Fdelete: " + Fdelete);
				#endif

				// Has the VSAM field_type changed -- ILLEGAL!
				if (strings_dont_match (field_type, Ftype))
				{
					string error_msg = "ISAM::FileConfig ERROR: FIELD TYPE has been changed in PARENT CLASS for field_name: " + field_name;
					Trace (error_msg);
					set_error (99999, error_msg);
				}

			}
		}
	}

// for debug purposes - log number of records in this file (if any)
	long num_recs = 0;

	if (FileOpen () == true)
		num_recs = FileNumRecords ();

	return (true);
}

//*******************************************************
//* Add non-index (normal) data field to Record Structure
//*******************************************************
bool ISAM::AddField (string field_name, string field_type)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::AddField called for: " + field_name + trace_file_name ());
	#endif

	reset_error_codes();

	if (check_new_field_parms (field_name, field_type) == false)
		return (false);

	// else ok to add new field to structure
	NUM_FIELDS_LOADED++;

	F_NAME       [NUM_FIELDS_LOADED] = field_name;
	F_TYPE       [NUM_FIELDS_LOADED] = field_type;
	F_INDEX      [NUM_FIELDS_LOADED] = "N";
	F_LENGTH	 [NUM_FIELDS_LOADED] = 0;
	F_IN_C_KEY   [NUM_FIELDS_LOADED] = false;

	#ifdef TRACER_CODE
		Trace ("ISAM::AddField DONE" + trace_file_name ());
	#endif

	return (true);
}

//***********************************************************************************************
//* Add INDEX Field to Record Structure
//*		--> FIELD TYPE must be a string!!
// ---------------------------------------------------------------------------------------------
//    NOTES:	single field secondary indexes have p_key appended to them automatically by VSAM
//				VSAM limit: index fields can only exist in first 157 fields of record format
//***********************************************************************************************
bool ISAM::AddIndexField (string field_name, string field_type, unsigned int max_key_length)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::AddIndexField called for: " + field_name + trace_file_name ());
	#endif

	reset_error_codes();

	if (check_new_field_parms (field_name, field_type) == false)
		return (false);

	if (field_type != "S")
	{
		set_error (FIELD_TYPE_INVALID , "Index Field must be of type STRING.");
		return (false);
	}

	if (P_KEY_FIELD_NAME.length() == 0)
	{
		set_error (INDEX_FIELD_ERROR , "Secondary Index Field cant be specified until after a primary key field has been specified.");
		return (false);
	}

	if (max_key_length + P_KEY_FIELD_LEN > VSAM_MAX_KEY_LENGTH)
	{
		set_error (INDEX_FIELD_ERROR , "Secondary Index Field length plus P_KEY_FIELD_LEN would exceed VSAM_MAX_KEY_LENGTH limit.");
		return (false);
	}

	if (NUM_FIELDS_LOADED >= VSAM_KEY_FIELD_NUM_LIMIT)
	{
		set_error (INDEX_FIELD_ERROR , "Adding Secondary Index Field would violate the VSAM_KEY_FIELD_NUM_LIMIT");
		return (false);
	}

	// else ok to add new field to structure
	NUM_FIELDS_LOADED++;

	F_NAME       [NUM_FIELDS_LOADED] = field_name;
	F_TYPE       [NUM_FIELDS_LOADED] = field_type;	// FIELD TYPE must be a string!!
	F_INDEX      [NUM_FIELDS_LOADED] = "Y";
	F_LENGTH     [NUM_FIELDS_LOADED] = max_key_length;
	F_IN_C_KEY   [NUM_FIELDS_LOADED] = false;

	#ifdef TRACER_CODE
		Trace ("ISAM::AddIndexField DONE" + trace_file_name ());
	#endif

	return (true);
}

//***********************************************************************************************
//* Add PRIMARY KEY Field to Record Structure
//* ---------------------------------------------------------------------------------------------
//*		** FIELD TYPE must be a string **
//* ---------------------------------------------------------------------------------------------
//* MUST have one and only one field desscribed at the Primray Key
//*		This is a LOGICAL field only - it is not a real field in the rec format
//*		but is used for priamry key initialization/tracking
//* ---------------------------------------------------------------------------------------------
//* --> BEST to always define the primary key first in the parent class
//***********************************************************************************************
bool ISAM::AddPrimaryKey (string field_name, string field_type, unsigned int max_key_length)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::AddPrimaryKey called for: " + field_name + trace_file_name ());
	#endif

	reset_error_codes();

	if (check_new_field_parms (field_name, field_type) == false)
		return (false);

	// must be a string
	if (field_type != "S")
	{
		set_error (FIELD_TYPE_INVALID , "Primary Key Field must be of type STRING.");
		return (false);
	}

	// must not already defined for this dataset
	if (P_KEY_FIELD_NAME.length() != 0)
	{
		set_error (P_KEY_FIELD_ERROR , "Primary Key Field already specified for this dataset.");
		return (false);
	}

	// key length must be smaller than limit
	// shoulf be as small as possible
	if (max_key_length > VSAM_MAX_KEY_LENGTH)
	{
		set_error (P_KEY_FIELD_ERROR , "Primary Key Field length would exceed VSAM_MAX_KEY_LENGTH limit.");
		return (false);
	}

	// else ok to add PRIMARY KEY field to structure
	F_NAME       [0] = field_name;
	F_TYPE       [0] = field_type;		// FIELD TYPE must be a string!!

	F_INDEX      [0] = "P";				// PRIMARY KEY!!
	F_LENGTH	 [0] = max_key_length;
	F_IN_C_KEY   [0] = false;

	// save these values NOW
	P_KEY_FIELD_LEN  = max_key_length;	// used later for secondary and compound index checks
	P_KEY_FIELD_NAME = field_name;		// important

	#ifdef TRACER_CODE
		Trace ("ISAM::AddPrimaryKey DONE" + trace_file_name ());
	#endif

	return (true);
}

//***********************************************************************************************
//* Add COMPUND INDEX Field to Record Structure
//* ---------------------------------------------------------------------------------------------
//*		** FIELD TYPE must be a string **
//* ---------------------------------------------------------------------------------------------
//*	NOTES: compound indexes have p_key appended to them automatically by VSAM
//*		   VSAM limit: index fields can only exist in first 157 fields of record format
//***********************************************************************************************
bool ISAM::AddCompoundKey (string index_name, string field_type, string field_names)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::AddCompoundKey called for index: " + index_name + ", with fields: " + field_names + trace_file_name ());
	#endif

	reset_error_codes();

	if (check_new_field_parms (index_name, field_type) == false)
		return (false);

	if (field_type != "S")
	{
		set_error (FIELD_TYPE_INVALID , "Compund Key Field must be of type STRING.");
		return (false);
	}

	if (P_KEY_FIELD_NAME.length() == 0)
	{
		set_error (C_KEY_FIELD_ERROR , "Compound Key Field cant be specified until after a primary key field has been specified.");
		return (false);
	}

	if (NUM_FIELDS_LOADED >= VSAM_KEY_FIELD_NUM_LIMIT)
	{
		set_error (C_KEY_FIELD_ERROR , "Adding Compound Key Field would violate the VSAM_KEY_FIELD_NUM_LIMIT.");
		return (false);
	}

	// else ok to add new field to structure
	NUM_FIELDS_LOADED++;

	F_NAME       [NUM_FIELDS_LOADED] = index_name;
	F_TYPE       [NUM_FIELDS_LOADED] = field_type;
	F_INDEX      [NUM_FIELDS_LOADED] = "C";			// COMPOUND KEY!!
	F_LENGTH	 [NUM_FIELDS_LOADED] = 0;			// NOT used with Compound Keys (calculated later)
	F_IN_C_KEY   [NUM_FIELDS_LOADED] = false;

	// now we need to detetmine which field numbers contain the field_names
	//		example: XC5.10:4.20,XC4:3.15
	//		means:   field 5.len10 + field 4.len20+ field 5.len 15

	string names_array [MAX_NUM_FIELDS + 1];

	unsigned int num_field_names = 0;
	unsigned int field_num       = 0;

	num_field_names = parse_field (field_names, names_array, "+", MAX_NUM_FIELDS);

	string key_len_str = "";

	for (unsigned int i = 1; i <= num_field_names; i++)
	{
		field_num = get_field_num (names_array [i]);
		if (field_num > 0)
		{
			// flag this field as part of compound key
			F_IN_C_KEY [field_num] = true;

			// build length spec string for key portion
			if (F_LENGTH [field_num] > 0)
				key_len_str = "." + int_to_string (F_LENGTH [field_num]);

			// now create format spec for C_Key field
			if (i == 1)
				F_PARTS [NUM_FIELDS_LOADED] = int_to_string (field_num) + key_len_str;
			else
				F_PARTS [NUM_FIELDS_LOADED] += ":" + int_to_string (field_num) + key_len_str;
		}
	}
	
	#ifdef TRACER_CODE
		Trace ("--> DONE: compound key format is: " + F_PARTS [NUM_FIELDS_LOADED]);
	#endif

	return (true);
}

//*******************************************************
//* Open the File FOR THIS OBJECT
//*******************************************************
bool ISAM::FileOpen (void)
{
	string file_name = PATH_NAME + FILE_NAME;

	#ifdef TRACER_CODE
		Trace ("ISAM::FileOpen called for file: " + file_name);
	#endif

	reset_error_codes();

	// see if file already opened
	if (FILE_OPEN == true)
	{
		#ifdef TRACER_CODE
			Trace ("--> file already opened, do nothing and return true");
		#endif

		return (true);
	}

	// short FAR PASCAL VsamOpen(                     // open a database (n/c)
	//                     LPSTR      lpFilespec,       // database file name
	//                     WORD       Omode,            // open mode, read only, read write, shared,
	//                     LPHANDLE   lphwmcb,          // returned, handle of open database
	//                     LPSTR      lpPassword,       // password of the vsam library
	//                     LPSTR      lpDbKey);         // database encryption key

	log_VSAM_call ("VsamOpen",
					file_name,
					int_to_string (READ_WRITE_SHARED),
					"&FILE_HANDLE",
					"VSAM_PASSWORD",
					"ENCRYPTION_KEY");

	short result = VsamOpen ((char *) file_name.c_str(), (WORD) READ_WRITE_SHARED, &FILE_HANDLE, (char *) VSAM_PASSWORD.c_str(), (char *) ENCRYPTION_KEY.c_str());

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamOpen OKAY. new FILE_HANDLE: " + long_to_string (FILE_HANDLE));
		#endif

		FILE_OPEN = true;
		return (true);
	}
	else
	{
		set_error (result , "VsamOpen Error: " + get_VSAM_error_msg (result));
		FILE_OPEN = false;
		return (false);
	}

	return (true);
}

//*******************************************************
//* CLOSE the File FOR THIS OBJECT
//*******************************************************
bool ISAM::FileClose (bool log_error = true)
{
	string file_name = PATH_NAME + FILE_NAME;

	#ifdef TRACER_CODE
		Trace ("FileClose called for file: " + file_name);
	#endif

	reset_error_codes();

	// see if file already opened
	if (FILE_OPEN != true)
	{
		if (log_error == true)
			set_error (FILE_CLOSE_ERROR , "ISAM file is not open. Skip Close ISAM file call.");

		return (false);
	}

	// now try to CLOSE the file
	//		short FAR PASCAL VsamClose(HANDLE    DatasetNumber);

	log_VSAM_call ("VsamClose", "HANDLE: " + long_to_string (FILE_HANDLE));

	short result = VsamClose (FILE_HANDLE);	// NOTICE that its NOT &FILE_HANDLE

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamClose OKAY");
		#endif

		FILE_OPEN   = false;
		FILE_HANDLE = 0;
		return (true);
	}
	else
	{
		set_error (result , "VsamClose Error: " + get_VSAM_error_msg (result));
		return (false);
	}

	return (true);
}

//*******************************************************
//* Must be called during destruction of object so the
//* memory allocated by VSAM for this record buffer is
//* released from the free store. 
//*******************************************************
bool ISAM::FreeVSAMBuffer (void)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::FreeVSAMBuffer called" + trace_file_name ());
	#endif

    if (VSAM_BUFFER)				// if pointer exists (not NULL)
    {
		free   (VSAM_BUFFER);		// C approach   - free the memory allocated to VSAM_BUFFER
		// delete (VSAM_BUFFER);	// C++ approach - free the memory allocated to VSAM_BUFFER
        VSAM_BUFFER = NULL;			// reset pointer to NULL

		#ifdef TRACER_CODE
			Trace ("--> VSAM rec buffer memory freed OK");
		#endif

		return (true);
    }
	else
	{
		#ifdef TRACER_CODE
			Trace ("--> VSAM rec buffer mem not currently allocated: skip FreeGstr call.");
		#endif

		return (false);
    }
}

//*******************************************************
//* returns the number of records in the file
//*******************************************************
long ISAM::FileNumRecords (void)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::FileNumRecords called." + trace_file_name ());
	#endif

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// typedef struct tagVSTATS
	//   {
	//     LONG   nrecords;
	//     WORD   gps_used;
	//     WORD   gps_unused;
	//     WORD   max_key_len;
	//     LONG   grp_size;
	//     WORD   init_alloc;
	//     WORD   cr_opts;
	//     WORD   free_prop_area;
	//     char   format[256];
	//     LONG   reserved;
	// 
	//   } VSTATS;

	VSTATS DBINFO;			// declare pointer to structure of type VSTATS

	reset_error_codes();

	log_VSAM_call ("VsamInfo", "HANDLE: " + long_to_string (FILE_HANDLE), "&DBINFO");

	short result = VsamInfo (FILE_HANDLE, &DBINFO);

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamInfo OKAY. Number of Records in File: " + long_to_string (DBINFO.nrecords));
		#endif

		return (DBINFO.nrecords);
	}
	else
	{
		set_error (result , "VsamInfo Error: " + get_VSAM_error_msg (result));
		return (0);
	}

}

//*******************************************************
//* DELETES the assigned VSAM Dataset 
//*******************************************************
bool ISAM::KillDataSet (void)
{

	#ifdef TRACER_CODE
		Trace ("ISAM::KillDataSet called for file: " + PATH_NAME + FILE_NAME + trace_file_name ());
	#endif


	// cant do this unless the file is CLOSED
	if (FILE_OPEN == true)
	{
		set_error (KILL_DATASET_ERROR, "VsamKill Not Allowed - ISAM file is currently open.");
		return (false);
	}

	// cant do this unless we have a file name - path can be blank, but not file name
	if (FILE_NAME.length() == 0)
	{
		set_error (KILL_DATASET_ERROR, "Dataset File Name to Kill is Missing or Invalid");
		return (false);
	}

	// load file path and name to delete into string for use below
	string file_to_delete = PATH_NAME + FILE_NAME;

	reset_error_codes();

	log_VSAM_call ("VsamKill", file_to_delete);

	//	short FAR PASCAL VsamKill(LPSTR  lpFilespec);

	short result = VsamKill ((char *) file_to_delete.c_str());

	// result processing 
	if (result == VIS_OK)
	{
		P_KEY_FIELD_VALUE = "";

		#ifdef TRACER_CODE
			Trace ("--> VsamKill OKAY");
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamKill  Error: " + get_VSAM_error_msg (result));
		return (false);
	}
}

//*******************************************************
//* called to set the value of a field in the VSAM_BUFFER
//* --> FOR STRING fields
//*******************************************************
bool ISAM::SetField (string field_name, string& field_value)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::SetField called for STRING field: " + field_name + ", field_value: " + field_value + trace_file_name ());
	#endif


	unsigned int field_num = 0;
	string field_type      = "";

	field_num = get_field_num (field_name);

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// invalid field name?
	if (field_num == 99999)
	{
		set_error (FIELD_NAME_INVALID , "Field Name Missing or Invalid.");
		return (false);
	}

	// are we updating logical PRIMARY KEY field?
	if (field_num == 0)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamStoreField OKAY: Primary Key Value Set.");
		#endif

		P_KEY_FIELD_VALUE = field_value;
		return (true);
	}

	// else set value of a regular field...
	field_type = get_vsam_field_type (F_TYPE [field_num]);

	// short FAR PASCAL VsamStoreField(               // store field data in a record (new)
	//                     HANDLE     DatasetNumber,    // handle of open database
	//                     LPGSTR    *lpRecord,         // pointer to record pointer
	//                     LPSTR      lpField,          // field to store into (name or number if %n)
	//                     WORD       wElement,         // NEW array element to store
	//                     LPSTR      lpFldType,        // data type of field (needed?)
	//                     void      *lpFieldData,      // pointer to field data
	//                     WORD       lFieldData);      // length of field data (if a string)

	// be sure to trim string length to MAX_STRING_LEN
	if (field_value.length() > MAX_STRING_LEN)
	{
		#ifdef TRACER_CODE
			Trace ("--> need to trim field down to MAX_STRING_LEN: " + int_to_string (MAX_STRING_LEN));
		#endif

		field_value = truncate_string (field_value, MAX_STRING_LEN);
	}

	reset_error_codes();

	WORD wElement = 0;

	log_VSAM_call ("VsamStoreField",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"&VSAM_BUFFER",
					"\"" + field_name + "\"",
					"wElement",						// NEW array element to store???
					field_type,
					field_value,
					int_to_string (field_value.length()));

	#ifdef TRACER_CODE
		Trace ("--> field_name being used: " + field_name);
	#endif


	short result = VsamStoreField (FILE_HANDLE, &VSAM_BUFFER, (char *) field_name.c_str(), (WORD) wElement, (char *) field_type.c_str(), (char *) field_value.c_str(), field_value.length());

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamStoreField OKAY");
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamStoreField Error: " + get_VSAM_error_msg (result));
		return (false);
	}

}

//*******************************************************
//* called to set the value of a field in the VSAM_BUFFER
//* --> FOR INTEGER fields
//*******************************************************
bool ISAM::SetField (string field_name, int& field_value)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::SetField called for INTEGER field: " + field_name + ", field_value: " + int_to_string (field_value) + trace_file_name ());
	#endif


	unsigned int field_num = 0;
	string field_type      = "";

	field_num = get_field_num (field_name);

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// invalid field name?
	if (field_num == 99999)
	{
		set_error (FIELD_NAME_INVALID , "Field Name Missing or Invalid.");
		return (false);
	}

	// CANT be updating logical PRIMARY KEY field as INTEGER FIELD?
	if (field_num == 0)
	{
		set_error (FIELD_TYPE_INVALID , "SetField call for Primary Key must be for field type of STRING.");
		return (false);
	}

	// else set value of a regular field...
	field_type = get_vsam_field_type (F_TYPE [field_num]);

	// short FAR PASCAL VsamStoreField(               // store field data in a record (new)
	//                     HANDLE     DatasetNumber,    // handle of open database
	//                     LPGSTR    *lpRecord,         // pointer to record pointer
	//                     LPSTR      lpField,          // field to store into (name or number if %n)
	//                     WORD       wElement,         // NEW array element to store
	//                     LPSTR      lpFldType,        // data type of field (needed?)
	//                     void      *lpFieldData,      // pointer to field data
	//                     WORD       lFieldData);      // length of field data (if a string)

	reset_error_codes();

	WORD wElement = 0;

	log_VSAM_call ("VsamStoreField",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"&VSAM_BUFFER",
					"\"" + field_name + "\"",
					"wElement",						// NEW array element to store???
					field_type,
					int_to_string(field_value),
					"0");

	// string field_num_str = "%" + int_to_string (field_num);

	#ifdef TRACER_CODE
		Trace ("--> field_name being used: " + field_name);
	#endif


	short result = VsamStoreField (FILE_HANDLE, &VSAM_BUFFER, (char *) field_name.c_str(), (WORD) wElement, (char *) field_type.c_str(), &field_value, 0);

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamStoreField OKAY");
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamStoreField Error: " + get_VSAM_error_msg (result));
		return (false);
	}

}
//*******************************************************
//* called to set the value of a field in the VSAM_BUFFER
//* --> FOR LONG fields
//*******************************************************
bool ISAM::SetField (string field_name, long& field_value)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::SetField called for LONG field: " + field_name + ", field_value: " + long_to_string (field_value) + trace_file_name ());
	#endif


	unsigned int field_num = 0;
	string field_type      = "";

	field_num = get_field_num (field_name);

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// invalid field name?
	if (field_num == 99999)
	{
		set_error (FIELD_NAME_INVALID , "Field Name Missing or Invalid.");
		return (false);
	}

	// CANT be updating logical PRIMARY KEY field as LONG FIELD?
	if (field_num == 0)
	{
		set_error (FIELD_TYPE_INVALID , "SetField call for Primary Key must be for field type of STRING.");
		return (false);
	}

	// else set value of a regular field...
	field_type = get_vsam_field_type (F_TYPE [field_num]);

	// short FAR PASCAL VsamStoreField(               // store field data in a record (new)
	//                     HANDLE     DatasetNumber,    // handle of open database
	//                     LPGSTR    *lpRecord,         // pointer to record pointer
	//                     LPSTR      lpField,          // field to store into (name or number if %n)
	//                     WORD       wElement,         // NEW array element to store
	//                     LPSTR      lpFldType,        // data type of field (needed?)
	//                     void      *lpFieldData,      // pointer to field data
	//                     WORD       lFieldData);      // length of field data (if a string)

	reset_error_codes();

	WORD wElement = 0;

	log_VSAM_call ("VsamStoreField",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"&VSAM_BUFFER",
					"\"" + field_name + "\"",
					"wElement",						// NEW array element to store???
					field_type,
					long_to_string(field_value),
					"0");

	#ifdef TRACER_CODE
		Trace ("--> field_name being used: " + field_name);
	#endif


	short result = VsamStoreField (FILE_HANDLE, &VSAM_BUFFER, (char *) field_name.c_str(), (WORD) wElement, (char *) field_type.c_str(), &field_value, 0);

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamStoreField OKAY");
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamStoreField Error: " + get_VSAM_error_msg (result));
		return (false);
	}

}

//*******************************************************
//* called to set the value of a field in the VSAM_BUFFER
//* --> FOR FLOAT fields
//*******************************************************
bool ISAM::SetField (string field_name, float& field_value)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::SetField called for FLOAT field: " + field_name + ", field_value: " + float_to_string (field_value) + trace_file_name ());
	#endif


	unsigned int field_num = 0;
	string field_type      = "";

	field_num = get_field_num (field_name);

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// invalid field name?
	if (field_num == 99999)
	{
		set_error (FIELD_NAME_INVALID , "Field Name Missing or Invalid.");
		return (false);
	}

	// CANT be updating logical PRIMARY KEY field as FLOAT FIELD?
	if (field_num == 0)
	{
		set_error (FIELD_TYPE_INVALID , "SetField call for Primary Key must be for field type of STRING.");
		return (false);
	}

	// else set value of a regular field...
	field_type = get_vsam_field_type (F_TYPE [field_num]);

	// short FAR PASCAL VsamStoreField(               // store field data in a record (new)
	//                     HANDLE     DatasetNumber,    // handle of open database
	//                     LPGSTR    *lpRecord,         // pointer to record pointer
	//                     LPSTR      lpField,          // field to store into (name or number if %n)
	//                     WORD       wElement,         // NEW array element to store
	//                     LPSTR      lpFldType,        // data type of field (needed?)
	//                     void      *lpFieldData,      // pointer to field data
	//                     WORD       lFieldData);      // length of field data (if a string)

	reset_error_codes();

	WORD wElement = 0;

	log_VSAM_call ("VsamStoreField",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"&VSAM_BUFFER",
					"\"" + field_name + "\"",
					"wElement",						// NEW array element to store???
					field_type,
					float_to_string(field_value),
					"0");

	#ifdef TRACER_CODE
		Trace ("--> field_name being used: " + field_name);
	#endif


	short result = VsamStoreField (FILE_HANDLE, &VSAM_BUFFER, (char *) field_name.c_str(), (WORD) wElement, (char *) field_type.c_str(), &field_value, 0);

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamStoreField OKAY");
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamStoreField Error: " + get_VSAM_error_msg (result));
		return (false);
	}

}

//*******************************************************
//* returns the number of fields loaded in file spec
//*******************************************************
unsigned int ISAM::FileGetNumFields (void)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::FileGetNumFields called to return: " + int_to_string (NUM_FIELDS_LOADED) + trace_file_name ());
	#endif

	return (NUM_FIELDS_LOADED);
}

//*******************************************************
//* gets the next record in the file based on the current
//* index open and pointer pos
//*------------------------------------------------------
//* optional parm:
//*		1) bool T/F - controls whether or not record buffer is loaded
//*******************************************************
bool ISAM::GetNextRecord (bool load_data = true)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::GetNextRecord called" + trace_file_name ());
	#endif


	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// make sure optional parm is valid
	string flag_str = "";
	if (load_data == true)
	{
		flag_str = "XGET_DATA";

		#ifdef TRACER_CODE
			Trace ("--> load_data = XGET_DATA");
		#endif

	}
	else
	{
		flag_str = "XNO_DATA";

		#ifdef TRACER_CODE
			Trace ("--> load_data = XNO_DATA");
		#endif

	}

//	short FAR PASCAL VsamGet(                      // read a record from database (n/c)
//                     HANDLE     DatasetNumber,    // handle of open database
//                     LPSTR      lpSecIndex,       // secondary index to use, (name or number if %n)
//                     WORD       Option,           // get option, next, previous, lookup, ...
//                     LPSTR      SrchKey,          // searh key if option is lookup
//                     LPSTR      SecKey,           // returned, secondary key if secindex not 0
//                     LPSTR      PriKey,           // returned, primary key
//                     LPGSTR    *lpRec);           // returned, pointer to record

	// be sure to unlock any record currently locked
	unlock (NO_LOG_ERROR);

	reset_error_codes();

	// storage for VSAM return keys
	char IndexKeyBuffer   [256];
	char PrimaryKeyBuffer [256];

	// perform operation
	short result = 0;

	log_VSAM_call ("VsamGet",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"\"" + CURR_INDEX_NAME + "\"",
					"XNEXT + " + flag_str,
					"\"\"",
					"IndexKeyBuffer",
					"PrimaryKeyBuffer",
					"&VSAM_BUFFER");

	#ifdef TRACER_CODE
		Trace ("--> CURR_INDEX_NAME being used: " + CURR_INDEX_NAME);
	#endif


	if (load_data == true)
		result = VsamGet (FILE_HANDLE, (char *) CURR_INDEX_NAME.c_str(), XNEXT + XGET_DATA, "", IndexKeyBuffer, PrimaryKeyBuffer, &VSAM_BUFFER);
	else
		result = VsamGet (FILE_HANDLE, (char *) CURR_INDEX_NAME.c_str(), XNEXT + XNO_DATA,  "", IndexKeyBuffer, PrimaryKeyBuffer, &VSAM_BUFFER);

	// clear these first,then load below if successful
	P_KEY_FIELD_VALUE = "";
	string index_key = "";

	// result processing (record data stored in VSAM_BUFFER right now)
	if (result == VIS_OK)
	{
		P_KEY_FIELD_VALUE.append (PrimaryKeyBuffer);
		index_key.append         (IndexKeyBuffer);

		#ifdef TRACER_CODE
			Trace ("--> VsamGet XNEXT OKAY");
			Trace ("----> PrimaryKey Returned.. " + P_KEY_FIELD_VALUE);
			Trace ("----> IndexKey Returned.... " + index_key);
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamGet XNEXT Error: " + get_VSAM_error_msg (result));
		return (false);
	}

}

//*******************************************************
//* gets the prev record in the file based on the current
//* index open and pointer pos
//*------------------------------------------------------
//* optional parm:
//*		1) bool T/F - controls whether or not record buffer is loaded
//*******************************************************
bool ISAM::GetPrevRecord (bool load_data = true)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::GetPrevRecord called" + trace_file_name ());
	#endif


	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// make sure optional parm is valid
	string flag_str = "";
	if (load_data == true)
	{
		flag_str = "XGET_DATA";

		#ifdef TRACER_CODE
			Trace ("--> load_data = XGET_DATA");
		#endif

	}
	else
	{
		flag_str = "XNO_DATA";

		#ifdef TRACER_CODE
			Trace ("--> load_data = XNO_DATA");
		#endif

	}

	// short FAR PASCAL VsamGet(HANDLE       DatasetNumber, 
	//                          WORD         IndexNum,
	//                          WORD         AccessOption, 
	//                          LPSTR        LookupStr,
	//                          LPSTR        IndexKeyReturned255, 
	//                          LPSTR        PrimaryKeyReturned255, 
	//                          LPGSTR      *VsamRecBuffer);

	// be sure to unlock any record currently locked
	unlock (NO_LOG_ERROR);

	reset_error_codes();

	// storage for VSAM return keys
	char IndexKeyBuffer   [256];
	char PrimaryKeyBuffer [256];

	// perform operation
	short result = 0;

	log_VSAM_call ("VsamGet",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"\"" + CURR_INDEX_NAME + "\"",
					"XPREVIOUS + " + flag_str,
					"\"\"",
					"IndexKeyBuffer",
					"PrimaryKeyBuffer",
					"&VSAM_BUFFER");

	#ifdef TRACER_CODE
		Trace ("--> CURR_INDEX_NAME being used: " + CURR_INDEX_NAME);
	#endif


	if (load_data == true)
		result = VsamGet (FILE_HANDLE, (char *) CURR_INDEX_NAME.c_str(), XPREVIOUS + XGET_DATA, "", IndexKeyBuffer, PrimaryKeyBuffer, &VSAM_BUFFER);
	else
		result = VsamGet (FILE_HANDLE, (char *) CURR_INDEX_NAME.c_str(), XPREVIOUS + XNO_DATA,  "", IndexKeyBuffer, PrimaryKeyBuffer, &VSAM_BUFFER);

	// clear these first,then load below if successful
	P_KEY_FIELD_VALUE = "";
	string index_key = "";

	// result processing (record data stored in VSAM_BUFFER right now)
	if (result == VIS_OK)
	{
		P_KEY_FIELD_VALUE.append (PrimaryKeyBuffer);
		index_key.append         (IndexKeyBuffer);

		#ifdef TRACER_CODE
			Trace ("--> VsamGet XPREV OKAY");
			Trace ("----> PrimaryKey Returned.. " + P_KEY_FIELD_VALUE);
			Trace ("----> IndexKey Returned.... " + index_key);
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamGet XPREV Error: " + get_VSAM_error_msg (result));
		return (false);
	}

}


//*******************************************************
//* Encrypts a text_string using a cypher_key
//* Returns encrypted hex_string result 
//*******************************************************
string ISAM::Encrypt (string text_str, string cypher_key)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::Encrypt called for: " + text_str + trace_file_name ());
		Trace ("              with cypher_key: " + cypher_key);
	#endif

	DWORD input_data_size = (DWORD) text_str.length();

	char result_buffer [input_data_size * 2];

	DWORD result_buffer_size = sizeof(result_buffer);

// short FAR PASCAL VsamEncrypt (
// 					 LPSTR  lpSourceData,		  // pointer to data block to be encrypted/decrypted
// 					 DWORD  lData,				  // length of data to encrypt/decrypt in bytes
// 					 LPSTR  lpEkey,				  // cypher_key (null terminated string)
// 					 LONG   CryptOpt,			  // 1 = Encrypt, 0= Decrypt
// 					 LPSTR  lpResult,             // Data Buffer to hold results
// 					 DWORD  lResult);			  // size of result buffer

	log_VSAM_call ( "VsamEncrypt", 
					"ASCII Text: " + text_str, 
					"Len: " + long_to_string (text_str.length()),
					"cypher_key",
					"1=ENCRYPT",
					"result_buffer",
					"Size: " + long_to_string (sizeof(result_buffer)));

	short result = VsamEncrypt ( (char *) text_str.c_str(), input_data_size, (char *) cypher_key.c_str(), 1, result_buffer, result_buffer_size);

	string hex_string = "";

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamEncrypt OKAY.");
		#endif

		hex_string.append(result_buffer, sizeof(result_buffer));

		return (hex_string);
	}
	else
	{
		set_error (result , "VsamEncrypt Call Error: " + get_VSAM_error_msg (result));
		return ("");
	}

}

//*******************************************************
//* Decrypts an encrypted hex_string using a cypher_key
//* Returns decrypted text_string result 
//*******************************************************
string ISAM::Decrypt (string hex_str, string cypher_key)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::Decrypt called for: " + hex_str + trace_file_name ());
		Trace ("              with cypher_key: " + cypher_key);
	#endif

	DWORD input_data_size = (DWORD) hex_str.length();

	char result_buffer [input_data_size / 2];

	DWORD result_buffer_size = sizeof(result_buffer);

// short FAR PASCAL VsamEncrypt (
// 					 LPSTR  lpSourceData,		  // pointer to data block to be encrypted/decrypted
// 					 DWORD  lData,				  // length of data to encrypt/decrypt in bytes
// 					 LPSTR  lpEkey,				  // cypher_key (null terminated string)
// 					 LONG   CryptOpt,			  // 1 = Encrypt, 0= Decrypt
// 					 LPSTR  lpResult,             // Data Buffer to hold results
// 					 DWORD  lResult);			  // size of result buffer

	log_VSAM_call ( "VsamEncrypt", 
					"HEX Text: " + hex_str, 
					"Len: " + long_to_string (hex_str.length()),
					"cypher_key",
					"0=DECRYPT",
					"result_buffer",
					"Size: " + long_to_string (sizeof(result_buffer)));

	short result = VsamEncrypt ( (char *) hex_str.c_str(),  input_data_size, (char *) cypher_key.c_str(), 0, result_buffer, result_buffer_size);

	string text_string = "";

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamEncrypt OKAY.");
		#endif

		text_string.append(result_buffer, sizeof(result_buffer));		// ONLY LOAD thru END of result BUFFER!

		return (text_string);
	}
	else
	{
		set_error (result , "VsamEncrypt Call Error: " + get_VSAM_error_msg (result));
		return ("");
	}

}


//*******************************************************
//* Fetch VSAM DB field name structure and returns the 
//* results in variables (by address). 
//*******************************************************
bool ISAM::GetFieldNameList (string &field_names, long &field_count)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::GetFieldNameList called" + trace_file_name ());
	#endif

//  short FAR PASCAL VsamEnumAttribValues(          // get list of attribute values - comma delimiter string
//                     HANDLE     DatasetNumber,    // handle of open database
//                     WORD       wOption,          // retrieval option (0=active, 1=deleted, 2=all)
//                     LPSTR      lpAttribName,     // VSAM attribute name to fetch
//                     LPSTR      lpFldNames,       // a buffer to hold field names (',' delimited)
//                     LPWORD     lpwNamesBufSize,  // size of FldNames buffer
//                     LPWORD     lpwNumFields);    // returned, actual number of field names

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return ("");
	}

	log_VSAM_call ("VsamEnumAttribValues", "HANDLE: " + long_to_string (FILE_HANDLE), "Fnam");

	WORD buffer_size   = sizeof(STRING_BUFFER);

	WORD active_fields  = 0;
//	WORD deleted_fields = 1;
//	WORD all_fields     = 2;

	WORD num_fields = 0;

	short result = VsamEnumAttribValues (FILE_HANDLE, active_fields, "Fnam", STRING_BUFFER, &buffer_size, &num_fields);

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamEnumAttribValues OKAY.");
		#endif

		field_names.append (STRING_BUFFER);
		field_count = (long) num_fields;

		return (true);
	}
	else
	{
		set_error (result , "VsamEnumAttribValues Error: " + get_VSAM_error_msg (result));
		return (false);
	}

}

//*******************************************************
//* Fetch VSAM DB field numbers and returns the 
//* results in variables (by address). 
//*******************************************************
bool ISAM::GetFieldNumList (string &field_numbers, long &field_count)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::GetFieldNumList called" + trace_file_name ());
	#endif

//  short FAR PASCAL VsamEnumAttribValues(          // get list of attribute values - comma delimiter string
//                     HANDLE     DatasetNumber,    // handle of open database
//                     WORD       wOption,          // retrieval option (0=active, 1=deleted, 2=all)
//                     LPSTR      lpAttribName,     // VSAM attribute name to fetch
//                     LPSTR      lpFldNames,       // a buffer to hold field names (',' delimited)
//                     LPWORD     lpwNamesBufSize,  // size of FldNames buffer
//                     LPWORD     lpwNumFields);    // returned, actual number of field names

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return ("");
	}

	log_VSAM_call ("VsamEnumAttribValues", "HANDLE: " + long_to_string (FILE_HANDLE), "Fnum");

	WORD buffer_size   = sizeof(STRING_BUFFER);

	WORD active_fields  = 0;
//	WORD deleted_fields = 1;
//	WORD all_fields     = 2;

	WORD num_fields = 0;

	short result = VsamEnumAttribValues (FILE_HANDLE, active_fields, "Fnum", STRING_BUFFER, &buffer_size, &num_fields);

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamEnumAttribValues OKAY.");
		#endif

		field_numbers.append (STRING_BUFFER);
		field_count = (long) num_fields;

		return (true);
	}
	else
	{
		set_error (result , "VsamEnumAttribValues Error: " + get_VSAM_error_msg (result));
		return (false);
	}
}



//*******************************************************
//* Returns the SINGLE Field Attribute Requested (PRIVATE)
//*******************************************************
string ISAM::GetFieldAttribute (string field_name, string attrib_name)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::GetFieldAttribute for field: " + field_name + ", attrib: " + attrib_name + " called." + trace_file_name ());
	#endif

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return ("");
	}

// attribute names are:
//	Field Name			"Fnam"		String Name of the field
//	Field Number		"Fnum"		 n -		Read Only
//	Field Class			"Fcls"		Class		Read Only
//	Field Type			"Ftyp"		[%,&,!,#,@,$,C] Read Only
//	Field Is an Index	"Find"		[True, False	]	Read Only
//	Field width			"Fwid"		field width in characters
//	Field Just			"Fjst"		[L,C,R]
//	Field Is  Deleted	"Fdel"		[True, False]	

// short FAR PASCAL VsamGetFieldAttribute(        // get an attribute for a field (new)
//                      HANDLE     DatasetNumber,    // handle of open database
//                      LPSTR      lpField,          // field to access (name or number if %n)
//                      LPSTR      lpAttribueName,   // name of field attribue to retrieve
//                      LPSTR      lpData,           // buffer to recieve attribute data
//                      LPWORD     lpwDataSize);     // size of buffer to hold field attribue

	WORD buffer_size   = sizeof(STRING_BUFFER);

	reset_error_codes();

	log_VSAM_call ("VsamGetFieldAttribute", "HANDLE: " + long_to_string (FILE_HANDLE), field_name, attrib_name);

	short result = VsamGetFieldAttribute (FILE_HANDLE, (char *) field_name.c_str(), (char *) attrib_name.c_str(), STRING_BUFFER, &buffer_size);

	string attribute_str = "";

	if (result == VIS_OK)
	{
		attribute_str.append (STRING_BUFFER);

		#ifdef TRACER_CODE
			Trace ("--> VsamGetFieldAttribute OKAY for: " + field_name);
		#endif

		return (attribute_str);
	}
	else
	{
		set_error (result , "VsamGetFieldAttribute Error: " + get_VSAM_error_msg (result));
		return ("");
	}

}


//************************************************************
//* TRACE PROGRAM STEPS - only done when file named in DEBUG_FILE_NAME 
//* exists which lets you debug a remote user's usage
//* of this code by simply creating the named debug file
//* in the local script directory on the user's web server
//************************************************************
void ISAM::Trace(string Tracer_text, bool func_header = false)
{	
	if (TRACER_MODE_CHECKED == false)	// not done yet
		check_debug_mode();				// are we in debug mode?

	if (TRACER_MODE == true && TRACER_OFF == false)	 // are we in debug mode?
	{
		if (func_header == true)
			Tracer_text = "FUNC CALL: " + Tracer_text;

		ofstream LogFile (TRACER_LOGFILE_NAME.c_str(), ios::app);		// opens and add to end of it	

		if (func_header == true)
			LogFile << "===========================================================" << "\n";

		LogFile << Tracer_text << " - " << get_current_time () << "\n";				// write line of text 

		if (func_header == true)
			LogFile << "===========================================================" << "\n";

		LogFile.close();								// close it 	
	}
}

//*******************************************************
//* sets pointer to BOF
//*******************************************************
bool ISAM::SetBOF (void)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::SetBOF called using CURR_INDEX_NAME: " + CURR_INDEX_NAME + trace_file_name ());
	#endif

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// short FAR PASCAL VsamBOF (HANDLE       DatasetNumber, 
	//                           WORD         SecIndex);

	reset_error_codes();

	log_VSAM_call ("VsamBOF", "HANDLE: " + long_to_string (FILE_HANDLE), "\"" + CURR_INDEX_NAME + "\"");

	// string index_num_str = "%" + int_to_string (CURR_INDEX_NUM);

	#ifdef TRACER_CODE
		Trace ("--> CURR_INDEX_NAME being used: " + CURR_INDEX_NAME);
	#endif

	//	short result = VsamBOF (FILE_HANDLE, (char *) index_num_str.c_str());
	short result = VsamBOF (FILE_HANDLE, (char *) CURR_INDEX_NAME.c_str());

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamBOF OKAY");
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamBOF Error: " + get_VSAM_error_msg (result));
		return (false);
	}

	return (true);
}

//*******************************************************
//* sets pointer to EOF
//*******************************************************
bool ISAM::SetEOF (void)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::SetEOF called using CURR_INDEX_NAME: " + CURR_INDEX_NAME + trace_file_name ());
	#endif

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// short FAR PASCAL VsamEOF (HANDLE       DatasetNumber, 
	//                           WORD         SecIndex);

	reset_error_codes();

	log_VSAM_call ("VsamEOF", "HANDLE: " + long_to_string (FILE_HANDLE), "\"" + CURR_INDEX_NAME + "\"");

	// string index_num_str = "%" + int_to_string (CURR_INDEX_NUM);

	#ifdef TRACER_CODE
		Trace ("--> CURR_INDEX_NAME being used: " + CURR_INDEX_NAME);
	#endif

//	short result = VsamEOF (FILE_HANDLE, (char *) index_num_str.c_str());
	short result = VsamEOF (FILE_HANDLE, (char *) CURR_INDEX_NAME.c_str());

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamEOF OKAY");
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamEOF Error: " + get_VSAM_error_msg (result));
		return (false);
	}

	return (true);
}

//*******************************************************
//* ADDS record loaded into VSAM_BUFFER into the file
//*******************************************************
bool ISAM::AddRecord (string primary_key, bool rec_must_not_exist = true)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::AddRecord called for p_key: " + primary_key + trace_file_name ());
	#endif


	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	if (primary_key.length() == 0)
	{
		set_error (ADD_RECORD_FAILED , "Primary Key Missing or Invalid.");
		return (false);
	}

	// short FAR PASCAL VsamPut(HANDLE       DatasetNumber, 
	//                          LPSTR        lpKey, 
	//                          LPGSTR       *lpRec, 
	//                          WORD         protmode);
	//
	// # define ADD_OR_REPLACE               0     // flag for write to allow add or replace
	// # define ADD_ONLY                     1     // flag for write to allow only adds
	// # define REPLACE_ONLY                 2     // flag for write to allow only replaces

	// first lock the file with this key
	if (lock (primary_key) == false)
		return (false);

	reset_error_codes();

	short result = 0;

	if (rec_must_not_exist == true)
	{
		log_VSAM_call ("VsamPut","HANDLE: " + long_to_string (FILE_HANDLE),primary_key,"VSAM_BUFFER","ADD_ONLY");
		result =        VsamPut  (FILE_HANDLE, (char *) primary_key.c_str(), VSAM_BUFFER, ADD_ONLY);
	}
	else
	{
		log_VSAM_call ("VsamPut","HANDLE: " + long_to_string (FILE_HANDLE),primary_key,"VSAM_BUFFER","ADD_OR_REPLACE");
		result =        VsamPut  (FILE_HANDLE, (char *) primary_key.c_str(), VSAM_BUFFER, ADD_OR_REPLACE);
	}

	// now unlock this file
	unlock ();

	// result processing
	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamPut ADD OKAY");
		#endif

		P_KEY_FIELD_VALUE = primary_key;
		return (true);
	}
	else
	{
		set_error (result , "VsamPut ADD Error: " + get_VSAM_error_msg (result));
		P_KEY_FIELD_VALUE = "";
		return (false);
	}

}

//*******************************************************
//* UPDATES record loaded into VSAM_BUFFER into the file
//*******************************************************
bool ISAM::UpdateRecord (string primary_key, bool rec_must_exist = false)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::UpdateRecord called for p_key: " + primary_key + trace_file_name ());
	#endif


	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	if (primary_key.length() == 0)
	{
		set_error (ADD_RECORD_FAILED , "Primary Key Missing or Invalid.");
		return (false);
	}

	// short FAR PASCAL VsamPut(HANDLE       DatasetNumber, 
	//                          LPSTR        lpKey, 
	//                          LPGSTR       *lpRec, 
	//                          WORD         protmode);
	//
	// # define ADD_OR_REPLACE               0     // flag for write to allow add or replace
	// # define ADD_ONLY                     1     // flag for write to allow only adds
	// # define REPLACE_ONLY                 2     // flag for write to allow only replaces

	// first lock the file with this key

	if (lock (primary_key) == false)
		return (false);

	reset_error_codes();

	short result = 0;

	if (rec_must_exist == true)
	{
		log_VSAM_call ("VsamPut","HANDLE: " + long_to_string (FILE_HANDLE),primary_key,"VSAM_BUFFER","REPLACE_ONLY");
		result =        VsamPut  (FILE_HANDLE, (char *) primary_key.c_str(), VSAM_BUFFER, REPLACE_ONLY);
	}
	else
	{
		log_VSAM_call ("VsamPut","HANDLE: " + long_to_string (FILE_HANDLE),primary_key,"VSAM_BUFFER","ADD_OR_REPLACE");
		result =        VsamPut  (FILE_HANDLE, (char *) primary_key.c_str(), VSAM_BUFFER, ADD_OR_REPLACE);
	}

	// now unlock this file
	unlock ();

	// result processing
	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamPut UPDATE OKAY");
		#endif

		P_KEY_FIELD_VALUE = primary_key;
		return (true);
	}
	else
	{
		set_error (result , "VsamPut UPDATE Error: " + get_VSAM_error_msg (result));
		P_KEY_FIELD_VALUE = "";
		return (false);
	}

}

//*******************************************************
//* DELETES a record by primary key
//*******************************************************
bool ISAM::DeleteRecord (string primary_key)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::DeleteRecord called for p_key: " + primary_key + trace_file_name ());
	#endif


	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// do we have a primary key?
	if (primary_key.length() == 0)
	{
		set_error (DELETE_RECORD_FAILED, "Primary Key String is Missing or Invalid");
		return (false);
	}

	// short FAR PASCAL VsamDelete(HANDLE    DatasetNumber, 
	//                             LPSTR     lpKey);

	reset_error_codes();

	// perform operation

	log_VSAM_call ("VsamDelete", "HANDLE: " + long_to_string (FILE_HANDLE),	primary_key);

	short result = VsamDelete (FILE_HANDLE, (char *) primary_key.c_str());

	// result processing 
	if (result == VIS_OK)
	{
		P_KEY_FIELD_VALUE = "";

		#ifdef TRACER_CODE
			Trace ("--> VsamDelete OKAY");
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamDelete  Error: " + get_VSAM_error_msg (result));
		return (false);
	}
}

//*******************************************************
//* sets file to use new INDEX field name
//* use PRIMARY to set to file's primary key field of 0
//*------------------------------------------------------
//* Calling this AUTO SETS rec POINTER TO BOF
//*******************************************************
bool ISAM::SetIndex (string new_index_name)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::SetIndex called to set index to: " + new_index_name + trace_file_name ());
	#endif

	// get field number associated with index
	unsigned int new_index_num = get_index_field_num (new_index_name);

	// error code is 99999	
	if (new_index_num == 99999)		// error?
	{
		set_error (INDEX_NAME_INVALID , "--> Invalid Index Field Name");
		return (false);
	}

	// else ok
	CURR_INDEX_NUM  = new_index_num;	// will be 0 (primary) or a valid index field num
	CURR_INDEX_NAME = new_index_name;

	// now move pointer to BOF when index changed
	// cant do this unless the file is opened
	if (FILE_OPEN == true)	
	{
		SetBOF();
	}

	// all done
	#ifdef TRACER_CODE
		Trace ("DONE: set new index OKAY and rec pointer now at BOF");
	#endif

	return (true);
}

//*******************************************************
//* called to GET the value of a STRING field 
//* in the VSAM_BUFFER
//*******************************************************
bool ISAM::GetField (string field_name, string& field_value)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::GetField called for STRING field: " + field_name + trace_file_name ());
	#endif


	unsigned int field_num = get_field_num (field_name);

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// invalid field name?
	if (field_num == 99999)
	{
		set_error (FIELD_NAME_INVALID , "Field Name Missing or Invalid.");
		return (false);
	}

	// load field type
	string field_type = get_vsam_field_type (F_TYPE [field_num]);

	// gets here - ok to get field from VSAM_BUFFER and return it
	// short FAR PASCAL VsamFetchField(HANDLE  DatasetNumber,
	//                                 LPGSTR *lpRecord,
	//                                 WORD    FldNo,
	//                                 LPSTR   FldType,
	//                                 void   *lpFieldData,
	//                                 WORD   *lplFieldData);

//		short FAR PASCAL VsamFetchField(            // fetch field data from a record (new)
//                     HANDLE     DatasetNumber,    // handle of open database
//                     LPGSTR     lpRecord,         // pointer to record
//                     LPSTR      lpField,          // field to fetch from (name or number if %n)
//                     WORD       wElement,         // NEW array element to fetch
//                     LPSTR      FldType,          // data type of field (needed?)
//                     void      *lpFieldData,      // pointer to buffer for fetched data
//                     WORD      *lplFieldData);    // length of buffer (only if a string)

	// pass in the size of the string buffer
	field_value = "";
	WORD buffer_size   = sizeof(STRING_BUFFER);

	reset_error_codes();

	// if just want primary key, return it
	if (field_num == 0)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamFetchField OKAY. Primary Key: " + P_KEY_FIELD_VALUE);
		#endif

		field_value = P_KEY_FIELD_VALUE;
		return (true);
	}

	WORD wElement = 0;

	// else let's get actual field calue from the buffer
	log_VSAM_call ("VsamFetchField",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"&VSAM_BUFFER",
					"\"" + field_name + "\"",
					"wElement",
					field_type,
					"&STRING_BUFFER",
					int_to_string (buffer_size));

	// string field_num_str = "%" + int_to_string (field_num);

	#ifdef TRACER_CODE
		Trace ("--> field_name being used: " + field_name);
	#endif


	short result = VsamFetchField (FILE_HANDLE, VSAM_BUFFER, (char *) field_name.c_str(), (WORD) wElement, (char *) field_type.c_str(), STRING_BUFFER, &buffer_size);

	if (result == VIS_OK)
	{
		field_value.append (STRING_BUFFER);

		#ifdef TRACER_CODE
			Trace ("--> VsamFetchField OKAY. Return " + field_name + ": " + field_value);
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamFetchField Error: " + get_VSAM_error_msg (result));
		return (false);
	}

}

//*******************************************************
//* called to GET the value of a INTEGER field 
//* in the VSAM_BUFFER
//*******************************************************
bool ISAM::GetField (string field_name, int& field_value)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::GetField called for INT field: " + field_name + trace_file_name ());
	#endif


	unsigned int field_num = get_field_num (field_name);

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// invalid field name?
	if (field_num == 99999)
	{
		set_error (FIELD_NAME_INVALID , "Field Name Missing or Invalid.");
		return (false);
	}

	// load field type
	string field_type = get_vsam_field_type (F_TYPE [field_num]);

	// declare field value and its type (INTEGER) for this function
	field_value = 0;

	// pass in the size 
	WORD field_size = sizeof(field_value);

//		short FAR PASCAL VsamFetchField(            // fetch field data from a record (new)
//                     HANDLE     DatasetNumber,    // handle of open database
//                     LPGSTR     lpRecord,         // pointer to record
//                     LPSTR      lpField,          // field to fetch from (name or number if %n)
//                     WORD       wElement,         // NEW array element to fetch
//                     LPSTR      FldType,          // data type of field (needed?)
//                     void      *lpFieldData,      // pointer to buffer for fetched data
//                     WORD      *lplFieldData);    // length of buffer (only if a string)

	reset_error_codes();

	WORD wElement = 0;

	// else let's get actual field calue from the buffer
	log_VSAM_call ("VsamFetchField",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"&VSAM_BUFFER",
					"\"" + field_name + "\"",
					"wElement",
					field_type,
					"&STRING_BUFFER",
					int_to_string (field_size));

	// string field_num_str = "%" + int_to_string (field_num);

	#ifdef TRACER_CODE
		Trace ("--> field_name being used: " + field_name);
	#endif


	short result = VsamFetchField (FILE_HANDLE, VSAM_BUFFER, (char *) field_name.c_str(), (WORD) wElement, (char *) field_type.c_str(), STRING_BUFFER, &field_size);

	if (result == VIS_OK)
	{
		// char string conversion to STRING class to int result field (the long way)
		string temp_str = "";
		temp_str.append (STRING_BUFFER);
		field_value = atoi(temp_str.c_str());

		#ifdef TRACER_CODE
			Trace ("--> VsamFetchField OKAY. Return " + field_name + ": " + int_to_string(field_value));
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamFetchField Error: " + get_VSAM_error_msg (result));
		return (false);
	}

}

//*******************************************************
//* called to GET the value of a LONG field 
//* in the VSAM_BUFFER
//*******************************************************
bool ISAM::GetField (string field_name, long& field_value)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::GetField called for LONG field: " + field_name + trace_file_name ());
	#endif


	unsigned int field_num = get_field_num (field_name);

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// invalid field name?
	if (field_num == 99999)
	{
		set_error (FIELD_NAME_INVALID , "Field Name Missing or Invalid.");
		return (false);
	}

	// load field type
	string field_type = get_vsam_field_type (F_TYPE [field_num]);

	// declare field value and its type (LONG) for this function
	field_value = 0;

	// pass in the size 
	WORD field_size = sizeof(field_value);

//		short FAR PASCAL VsamFetchField(            // fetch field data from a record (new)
//                     HANDLE     DatasetNumber,    // handle of open database
//                     LPGSTR     lpRecord,         // pointer to record
//                     LPSTR      lpField,          // field to fetch from (name or number if %n)
//                     WORD       wElement,         // NEW array element to fetch
//                     LPSTR      FldType,          // data type of field (needed?)
//                     void      *lpFieldData,      // pointer to buffer for fetched data
//                     WORD      *lplFieldData);    // length of buffer (only if a string)

	reset_error_codes();

	WORD wElement = 0;

	log_VSAM_call ("VsamFetchField",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"&VSAM_BUFFER",
					"\"" + field_name + "\"",
					"wElement",
					field_type,
					"&STRING_BUFFER",
					int_to_string (field_size));

	// string field_num_str = "%" + int_to_string (field_num);

	#ifdef TRACER_CODE
		Trace ("--> field_name being used: " + field_name);
	#endif

	short result = VsamFetchField (FILE_HANDLE, VSAM_BUFFER, (char *) field_name.c_str(), (WORD) wElement, (char *) field_type.c_str(), STRING_BUFFER, &field_size);

	if (result == VIS_OK)
	{
		// char string conversion to STRING class to long result field (the long way)
		string temp_str = "";
		temp_str.append (STRING_BUFFER);
		field_value = atol(temp_str.c_str());

		#ifdef TRACER_CODE
			Trace ("--> VsamFetchField OKAY. Return " + field_name + ": " + long_to_string(field_value));
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamFetchField Error: " + get_VSAM_error_msg (result));
		return (false);
	}

}

//*******************************************************
//* called to GET the value of a FLOAT field 
//* in the VSAM_BUFFER
//*******************************************************
bool ISAM::GetField (string field_name, float& field_value)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::GetField called for FLOAT field: " + field_name + trace_file_name ());
	#endif


	unsigned int field_num = get_field_num (field_name);

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// invalid field name?
	if (field_num == 99999)
	{
		set_error (FIELD_NAME_INVALID , "Field Name Missing or Invalid.");
		return (false);
	}

	// load field type
	string field_type = get_vsam_field_type (F_TYPE [field_num]);

	// declare field value and its type (LONG) for this function
	field_value = 0;

	// pass in the size 
	WORD field_size = sizeof(field_value);

//		short FAR PASCAL VsamFetchField(            // fetch field data from a record (new)
//                     HANDLE     DatasetNumber,    // handle of open database
//                     LPGSTR     lpRecord,         // pointer to record
//                     LPSTR      lpField,          // field to fetch from (name or number if %n)
//                     WORD       wElement,         // NEW array element to fetch
//                     LPSTR      FldType,          // data type of field (needed?)
//                     void      *lpFieldData,      // pointer to buffer for fetched data
//                     WORD      *lplFieldData);    // length of buffer (only if a string)

	reset_error_codes();

	WORD wElement = 0;

	log_VSAM_call ("VsamFetchField",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"&VSAM_BUFFER",
					"\"" + field_name + "\"",
					"wElement",
					field_type,
					"&STRING_BUFFER",
					int_to_string (field_size));

	// string field_num_str = "%" + int_to_string (field_num);

	#ifdef TRACER_CODE
		Trace ("--> field_name being used: " + field_name);
	#endif


	short result = VsamFetchField (FILE_HANDLE, VSAM_BUFFER, (char *) field_name.c_str(), (WORD) wElement, (char *) field_type.c_str(), STRING_BUFFER, &field_size);

	if (result == VIS_OK)
	{
		// char string conversion to STRING class to float result field (the long way)
		string temp_str = "";
		temp_str.append (STRING_BUFFER);
		field_value = atof(temp_str.c_str());

		#ifdef TRACER_CODE
			Trace ("--> VsamFetchField OKAY. Return " + field_name + ": " + float_to_string(field_value));
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamFetchField Error: " + get_VSAM_error_msg (result));
		return (false);
	}

}

//*******************************************************
//* sets file pointer to a specific location in the file
//* based on index_str and current index file open
//*------------------------------------------------------
//* optional parms:
//*		1) index key string to set the pointer to
//*		2) bool T/F - controls whether or not record buffer is loaded
//*******************************************************
bool ISAM::FindRecord (string index_str, bool load_data = true)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::FindRecord called for index: " + index_str + trace_file_name ());
	#endif


	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// make sure optional parm is valid
	string flag_str = "";
	if (load_data == true)
	{
		flag_str = "XGET_DATA";

		#ifdef TRACER_CODE
			Trace ("--> load_data = XGET_DATA");
		#endif

	}
	else
	{
		flag_str = "XNO_DATA";

		#ifdef TRACER_CODE
			Trace ("--> load_data = XNO_DATA");
		#endif

	}

	// short FAR PASCAL VsamGet(HANDLE       DatasetNumber, 
	//                          WORD         IndexNum,
	//                          WORD         AccessOption, 
	//                          LPSTR        LookupStr,
	//                          LPSTR        IndexKeyReturned255, 
	//                          LPSTR        PrimaryKeyReturned255, 
	//                          LPGSTR      *VsamRecBuffer);

	// be sure to unlock any record currently locked
	unlock (NO_LOG_ERROR);

	reset_error_codes();

	// storage for VSAM return keys
	char IndexKeyBuffer   [256];
	char PrimaryKeyBuffer [256];

	// perform operation
	short result = 0;

	log_VSAM_call ("VsamGet",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"\"" + CURR_INDEX_NAME + "\"",
					"XLOOKUP + " + flag_str,
					index_str,
					"IndexKeyBuffer",
					"PrimaryKeyBuffer",
					"&VSAM_BUFFER");

	#ifdef TRACER_CODE
		Trace ("--> CURR_INDEX_NAME being used: " + CURR_INDEX_NAME);
	#endif


	if (load_data == true)
		result = VsamGet (FILE_HANDLE, (char *) CURR_INDEX_NAME.c_str(), XLOOKUP + XGET_DATA, (char *) index_str.c_str(), IndexKeyBuffer, PrimaryKeyBuffer, &VSAM_BUFFER);
	else
		result = VsamGet (FILE_HANDLE, (char *) CURR_INDEX_NAME.c_str(), XLOOKUP + XNO_DATA,  (char *) index_str.c_str(), IndexKeyBuffer, PrimaryKeyBuffer, &VSAM_BUFFER);

	// clear these first,then load below if successful
	P_KEY_FIELD_VALUE = "";
	string index_key = "";

	// result processing (record data stored in VSAM_BUFFER right now)
	if (result == VIS_OK)
	{
		P_KEY_FIELD_VALUE.append (PrimaryKeyBuffer);
		index_key.append         (IndexKeyBuffer);

		#ifdef TRACER_CODE
			Trace ("--> VsamGet LOOKUP OKAY");
			Trace ("----> PrimaryKey Returned.. " + P_KEY_FIELD_VALUE);
			Trace ("----> IndexKey Returned.... " + index_key);
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamGet XLOOKUP Error: " + get_VSAM_error_msg (result));
		return (false);
	}

}

//*******************************************************
//* gets a record by primary key (index field 0)
//*------------------------------------------------------
//* optional parms:
//*		1) primary key to get
//*		2) bool T/F - controls whether or not record buffer is loaded
//*******************************************************
bool ISAM::GetRecord (string primary_key, bool load_data = true)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::GetRecord called for p_key: " + primary_key + trace_file_name ());
	#endif


	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// make sure optional parm is valid
	string flag_str = "";
	if (load_data == true)
	{
		flag_str = "XGET_DATA";

		#ifdef TRACER_CODE
			Trace ("--> load_data = XGET_DATA");
		#endif

	}
	else
	{
		flag_str = "XNO_DATA";

		#ifdef TRACER_CODE
			Trace ("--> load_data = XNO_DATA");
		#endif

	}

	// short FAR PASCAL VsamGet(HANDLE       DatasetNumber, 
	//                          WORD         IndexNum,
	//                          WORD         AccessOption, 
	//                          LPSTR        LookupStr,
	//                          LPSTR        IndexKeyReturned255, 
	//                          LPSTR        PrimaryKeyReturned255, 
	//                          LPGSTR      *VsamRecBuffer);

	// be sure to unlock any record currently locked
	unlock (NO_LOG_ERROR);

	reset_error_codes();

	// storage for VSAM return keys
	char IndexKeyBuffer   [256];
	char PrimaryKeyBuffer [256];

	// perform operation
	short result = 0;

	log_VSAM_call ("VsamGet",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"P_KEY FLD 0",
					"XLOOKUP + " + flag_str,
					primary_key,
					"IndexKeyBuffer",
					"PrimaryKeyBuffer",
					"&VSAM_BUFFER");

	if (load_data == true)
		result = VsamGet (FILE_HANDLE, "%0", XLOOKUP + XGET_DATA, (char *) primary_key.c_str(), IndexKeyBuffer, PrimaryKeyBuffer, &VSAM_BUFFER);
	else
		result = VsamGet (FILE_HANDLE, "%0", XLOOKUP + XNO_DATA,  (char *) primary_key.c_str(), IndexKeyBuffer, PrimaryKeyBuffer, &VSAM_BUFFER);

	// clear these first,then load below if successful
	P_KEY_FIELD_VALUE = "";
	string index_key = "";

	// result processing (record data stored in VSAM_BUFFER right now)
	if (result == VIS_OK)
	{
		P_KEY_FIELD_VALUE.append (PrimaryKeyBuffer);
		index_key.append         (IndexKeyBuffer);

		#ifdef TRACER_CODE
			Trace ("--> VsamGet LOOKUP OKAY");
			Trace ("----> PrimaryKey Returned.. " + P_KEY_FIELD_VALUE);
			Trace ("----> IndexKey Returned.... " + index_key);
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamGet XLOOKUP Error: " + get_VSAM_error_msg (result));
		return (false);
	}

}























































//********************************************************
//** converts integer variable to string - no padding
//********************************************************
string ISAM::int_to_string(const int& value_in)
{
	string str_out  = "";
	int result_size = 0;
	char buffer[75] = {0};

	result_size = sprintf(buffer, "%d", value_in);

	str_out = buffer;

	return(str_out);
}

//********************************************************
//** converts VSAM HANDLE variable to string
//********************************************************
string ISAM::long_to_string(const HANDLE& handle_value_in)
{
	long value_in = (LONG) handle_value_in;

	string str_out  = "";
	int result_size = 0;
	char buffer[75] = {0};

	result_size = sprintf(buffer, "%ld", value_in);

	str_out = buffer;
	
	return(str_out);
}

//********************************************************
//** converts long variable to string
//********************************************************
string ISAM::long_to_string(const long& value_in)
{
	string str_out  = "";
	int result_size = 0;
	char buffer[75] = {0};

	result_size = sprintf(buffer, "%ld", value_in);

	str_out = buffer;
	
	return(str_out);
}

//********************************************************
//** converts float variable to string (2 decimals)
//********************************************************
string ISAM::float_to_string(const float& value_in)
{
	string str_out  = "";
	int result_size = 0;
	char buffer[75] = {0};

	result_size = sprintf(buffer, "%.2f", value_in);

	str_out = buffer;
	
	return(str_out);
}

//*******************************************************
//* checks to see if strings are an exact match in
//* content and length - coverts and trims boths strings
//* to upper case before comparision
//* CASE INSENSITIVE
//*******************************************************
bool ISAM::strings_match (string str_1, string str_2)
{
	int MATCH = 0;

	str_1 = trim(ucase(str_1));
	str_2 = trim(ucase(str_2));
	
	if (str_1.compare(str_2) == MATCH 
		&& str_1.length() == str_2.length())
	{
		return (true);
	}
	else
	{
		return (false);
	}
}

//*******************************************************
//* checks to see if strings are an NOT exact match in
//* content and length - coverts and trims boths strings
//* to upper case before comparision
//* CASE INSENSITIVE
//*******************************************************
bool ISAM::strings_dont_match (string str_1, string str_2)
{
	int MATCH = 0;

	str_1 = trim(ucase(str_1));
	str_2 = trim(ucase(str_2));

	if (str_1.compare(str_2) != MATCH 
		|| str_1.length() != str_2.length())
	{
		return (true);
	}
	else
	{
		return (false);
	}
}

//****************************************************************
//** called to trim a string of BOTH leading and trailing blanks
//****************************************************************
string ISAM::trim(string string_in)
{
	string_in = ltrim(string_in);
	string_in = rtrim(string_in);

	return (string_in);
}

//****************************************************************
//** called to LEFT trim a string
//****************************************************************
string ISAM::ltrim(string string_in)
{
	int length = string_in.length();	// define and get length

	if (length == 0)					// nothing to trim
		return (string_in);				// return string unchanged

// erase all leading blanks 
	while (string_in[0] == ' ')			// blank in first position?
	{
		string_in.erase(0, 1);			// erase first position

		if (string_in.length()== 0)		// if now an empty string
			break;						// exit from this while loop
	}

	return (string_in);					// return results
}

//****************************************************************
//** called to RIGHT trim a string
//****************************************************************
string ISAM::rtrim(string string_in)
{
	int length = string_in.length();	// define and get length

	if (length == 0)					// nothing to trim
		return (string_in);				// return string unchanged

// erase all ending blanks - length must be correct before this loop
	while (string_in[length - 1] == ' ')	// blank in LAST position?
	{
		string_in.erase(length - 1, 1);		// erase LAST position

		length = string_in.length();		// get new length
		if (length == 0)					// if empty string
			break;							// exit from this while loop
	}

	return (string_in);
}

//********************************************************
//** converts string "str_in" to UPPER case string
//********************************************************
string ISAM::ucase(string str_in)
{
	int length = str_in.length();
	int i = 0;

	while (str_in[i] != '\0' && i < length)
	{
		if (isalpha(str_in[i]))
			str_in[i] = toupper(str_in[i]);
		i++;
	}
	return (str_in);
}

//************************************************************
//* check to see if we need to log steps (in debug mode?) 
//************************************************************
void ISAM::check_debug_mode(void)
{
	TRACER_MODE = false;				// reset flag	

	FILE *file_ptr = 0;				// declare and reset ptr to file handle
	file_ptr = fopen(DEBUG_FILE_NAME.c_str(), "r");	// see if we can open file for read

	if (file_ptr != NULL)			// "debug" file exists?
	{
		TRACER_MODE = true;				// set on when there - else off
		fclose(file_ptr);				// close file - we are done with it
	}
	TRACER_MODE_CHECKED = true;		// been done once now
}

//************************************************************
//* Can be called to clear named logfile if debug_mode true
//************************************************************
void ISAM::clear_log_file(void)
{
	check_debug_mode();				// are we in debug mode?

	if (TRACER_MODE == true)	
		ofstream LogFile (TRACER_LOGFILE_NAME.c_str(), ios::out);
}


//*******************************************************
//* log's current time into the Tracer file
//*******************************************************
string ISAM::get_current_time (void)
{
	time_t t_sys_time;				// define global time variable
	time (&t_sys_time);				// load time in seconds since epoch - NOTE uses data type time_t

	long   sys_time     = static_cast<long>(t_sys_time);	// conv data to type long from time_t type
	string sys_time_str = long_to_string(sys_time);			// convert to string as "seconds" since epoch

	return (sys_time_str);
}

//*******************************************************
//* log's current time into the Tracer file
//*******************************************************
void ISAM::log_current_time (void)
{
	LOG_TIME_COUNT++;

	#ifdef TRACER_CODE
		Trace ("TIME: " + get_current_time() + " (ISAM: " + long_to_string (LOG_TIME_COUNT) + ")");
	#endif
}


//*******************************************************
//* Checks to be sure new field adheres to ISAM class's
//* limits/logic/types
//*******************************************************
bool ISAM::check_new_field_parms (string field_name, string field_type)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::check_new_field_parms for: " + field_name + ", type: " + field_type + trace_file_name ());
	#endif

	// already full?
	if (NUM_FIELDS_LOADED >= MAX_NUM_FIELDS)
	{
		set_error (TOO_MANY_FIELDS, "Too Many Fields: Would Exceed ISAM::MAX_NUM_FIELDS");
		return (false);
	}

	// field name not blank
	if (field_name.length() == 0)
	{
		set_error (FIELD_NAME_INVALID , "Field Name Missing.");
		return (false);
	}

	// no field name cant be "PRIMARY"
	if (strings_match (field_name, PRIMARY_KEY))
	{
		set_error (FIELD_NAME_INVALID , "Field Name PRIMARY reserved for internal use,");
		return (false);
	}

	// field type valid?
	if (field_type.length() == 0)
	{
		set_error (FIELD_TYPE_INVALID , "Field Type Missing.");
		return (false);
	}

	// type of field S=string, L=long, I=Int, F=float
	if (field_type != "S" && field_type != "L" && field_type != "I" && field_type != "F")
	{
		set_error (FIELD_TYPE_INVALID , "Field Type Invalid.");
		return (false);
	}

	// make sure this new field does not already exist
	if (get_field_num (field_name) != 99999)		// if not a new field?
	{
		set_error (FIELD_TYPE_INVALID , "Duplicate Field or Index Name Detected");
		return (false);
	}

	// if gets here - all ok
	return (true);

}
//***********************************************************************
//* Resets error codes - usually called at the start of internal function
//***********************************************************************
void ISAM::reset_error_codes (void)
{
	error   = false;
	err_num = 0;
	err_msg = "";

	return;
}

//***********************************************************************
//* takes string that is delimited by ascII character passed in
//* and loads up a passed in reference to an array of results 1 - xx
//***********************************************************************
unsigned int ISAM::parse_field (string data_field, string result_array [], string delimiter, unsigned int max_fields)
{
	#ifdef TRACER_CODE
		Trace ("parse_field called for field: " + data_field + trace_file_name ());
	#endif

	reset_error_codes();

	// first clear result array
	for (unsigned int i = 1; i <= max_fields; i++)
	{
		result_array [i] = "";
	}

	// if nothing to parse, exit with 0 num of items
	if (data_field.length() == 0)
	{	
		#ifdef TRACER_CODE
			Trace ("--> ERROR: data_field is blank - exit, nothing to parse");
		#endif

		return (0);
	}

	// else ok to parse
	unsigned int num_fields = 0;

	unsigned int last_delimiter_pos = 0;
	unsigned int next_delimiter_pos = 0;

	string next_value = "";

	// parse loop
	#ifdef TRACER_CODE
		Trace ("--> start parse loop");
	#endif

	while (num_fields < max_fields)
	{
		next_delimiter_pos = data_field.find (delimiter, last_delimiter_pos+1);

		if (next_delimiter_pos != string::npos && next_delimiter_pos != last_delimiter_pos)
		{
			if (last_delimiter_pos != 0)
				next_value = data_field.substr (last_delimiter_pos+1, ((next_delimiter_pos-1) - last_delimiter_pos));
			else
				next_value = data_field.substr (last_delimiter_pos, ((next_delimiter_pos) - last_delimiter_pos));

			num_fields++;
			last_delimiter_pos = next_delimiter_pos;
			result_array [num_fields] = next_value;
		}
		else
		{
			if (last_delimiter_pos < data_field.length())
			{
				next_delimiter_pos = data_field.length();
				if (last_delimiter_pos != 0)
					next_value = data_field.substr (last_delimiter_pos+1, ((next_delimiter_pos-1) - last_delimiter_pos));
				else
					next_value = data_field.substr (last_delimiter_pos, ((next_delimiter_pos) - last_delimiter_pos));

				num_fields++;
				last_delimiter_pos = data_field.length();
				result_array [num_fields] = next_value;
			}
			else
				break;
		}
	}

	// when all said and done, there MUST be a primary key defined above somewhere
	if (P_KEY_FIELD_NAME.length() == 0)
	{
		set_error (P_KEY_MISSING , "No Primary Key Specified for Dataset.");
	}

	// return number of elements parsed
	return (num_fields);
}

//*******************************************************
//* looks through the arrays to find the field number
//* 	associated with a specific field name
//*     RETURNS 99999 on error
//*******************************************************
unsigned int ISAM::get_field_num (string field_name)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::get_field_num for field: " + field_name + trace_file_name ());
	#endif

	// are we looking for primary key field?
	if (strings_match (field_name, "PRIMARY") || strings_match (field_name, P_KEY_FIELD_NAME))
	{
		if (P_KEY_FIELD_NAME.length() != 0)		// primary key field specified in file spec?
		{
			#ifdef TRACER_CODE
				Trace ("--> Matches p_key field num: 0");
			#endif

			return (0);
		}
		else
		{
			#ifdef TRACER_CODE
				Trace ("--> Primary Key Field not Specified for this file. Return Code 99999");
			#endif

			return (99999);			// 99999 = not found
		}
	}

	// else see if we have a regular/index field by this name
	for (unsigned int i = 1; i <= NUM_FIELDS_LOADED; i++)
	{
		if (strings_match (field_name, F_NAME [i]))
			return (i);
	}

	return (99999);			// 99999 = not found
}

//*******************************************************
//* looks through the arrays to find the INDEX field number
//* 	associated with a specific field name
//*     RETURNS 99999 on error
//*******************************************************
unsigned int ISAM::get_index_field_num (string index_name)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::get_index_field_num for index: " + index_name + trace_file_name ());
	#endif

	// are we looking for primary key field?
	if (strings_match (index_name, "PRIMARY") || strings_match (index_name, P_KEY_FIELD_NAME))
	{
		if (P_KEY_FIELD_NAME.length() != 0)		// primary key field specified in file spec?
		{
			#ifdef TRACER_CODE
				Trace ("--> Matches p_key field num: 0");
			#endif

			return (0);
		}
		else
		{
			#ifdef TRACER_CODE
				Trace ("--> Primary Key Field not Specified for this file. Return Code 99999");
			#endif

			return (99999);		// not found
		}
	}

	// else see if we have a secondary or compound index by this name
	for (unsigned int i = 1; i <= NUM_FIELDS_LOADED; i++)
	{
		if (strings_match (index_name, F_NAME [i]))
		{
			if (F_INDEX [i] == "Y")
			{
				#ifdef TRACER_CODE
					Trace ("--> Matches Secondary Key Field Num: " + int_to_string (i));
				#endif

				return (i);
			}

			if (F_INDEX [i] == "C")
			{
				#ifdef TRACER_CODE
					Trace ("--> Matches Compound Key Field Num: " + int_to_string (i));
				#endif

				return (i);
			}
		}
	}

	// else this is not an index name
	#ifdef TRACER_CODE
		Trace ("--> Invalid Index Field Name. Return Code 99999");
	#endif

	return (99999);		// not found
}

//*******************************************************
//* Based on BSI field type (S, L, I, F), returns the 
//* correct VSAM field type
//*******************************************************
string ISAM::get_vsam_field_type (string bsi_field_type)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::get_vsam_field_type for bsi_field_type: " + bsi_field_type + trace_file_name ());
	#endif

	if (bsi_field_type == "S")	// string
		return ("$");

	if (bsi_field_type == "I")	// int
		return ("%");

	if (bsi_field_type == "L")	// long
		return ("&");

	if (bsi_field_type == "F")	// float
		return ("#");

	// gets here = invalid
	return ("$");	// default VSAM type is string

}

//********************************************************
//  checks to see if ISAM file exists
//--------------------------------------------------------
//  Example:  checks for existance of" Qfiles/Quser.vod
//
//  if (does_isam_file_exist ("Qfiles/", "Quser") == true)
//
//--------------------------------------------------------
// isam_file_name must be something like "Qsession" with 
// no path or file suffix.
//
// isam_dir_path cant be blank and should contain the 
// trailing slash
//
// We find it best to place all datafiles in a sub-dir
// underneath of the script directory. This allows for 
// better organization. Also, the file sub-dir could then 
// easily be protected from web access with .htaccess and 
// this approach also makes database backup easier. 
//********************************************************
bool ISAM::does_isam_file_exist (string isam_dir_path, string isam_file_name)
{
	if (isam_dir_path.length() == 0)	
		isam_dir_path = "./";		// make it the local directory

	#ifdef TRACER_CODE
		Trace ("ISAM::does_isam_file_exist called for CURR DIR file: " + isam_dir_path + isam_file_name);
		Trace ("--> Check for file manually by reading dir");
		Trace ("--> check dir: " + isam_dir_path + " - for file: " + isam_file_name);
	#endif

	if (isam_file_name.length() == 0)
	{
		#ifdef TRACER_CODE
			Trace ("--> isam_file_name is empty or invalid");
		#endif
		return(false);
	}

	if (isam_dir_path.length() == 0)
	{
		#ifdef TRACER_CODE
			Trace ("--> isam_dir_path is empty or invalid");
		#endif
		return(false);
	}

	// dir read and test
	bool file_ok = false;		// reset flag

	DIR *dp;						// declare pointer to a DIR 
	struct dirent *entry;			// declare a pointer to a DIRENT struct
	struct stat statbuf;			// create a status buffer structure
	
	string open_path = remove_dir_trailing_slash (isam_dir_path);

	dp = opendir(open_path.c_str());	// open DIR stream

	if (dp == NULL)						// bad path
	{
		#ifdef TRACER_CODE
			Trace ("--> isam_dir_path not found - File Cant Exist");
		#endif

		return(false);
	}

	while ((entry = readdir(dp)) != NULL)	// read directory entries
	{
		stat(entry->d_name,&statbuf);		// load entry result in status buffer
		string name = entry->d_name;		// load file name as a string

		#ifdef TRACER_CODE
			Trace ("--> checking file: " + name + " for match...");
		#endif

		if (strings_match (name, isam_file_name + ".vod"))	// dont forget the suffix!!!!
		{
			#ifdef TRACER_CODE
				Trace ("****************************");
				Trace ("----> ISAM .vod FILE EXISTS");
				Trace ("****************************");
			#endif

			file_ok = true;
			break;
		}
		if (strings_match (name, isam_file_name + ".vom"))	// dont forget the suffix!!!!
		{
			#ifdef TRACER_CODE
				Trace ("****************************");
				Trace ("----> ISAM .vom FILE EXISTS");
				Trace ("****************************");
			#endif

			file_ok = true;
			break;
		}
		if (strings_match (name, isam_file_name + ".vol"))	// dont forget the suffix!!!!
		{
			#ifdef TRACER_CODE
				Trace ("****************************");
				Trace ("----> ISAM .vol FILE EXISTS");
				Trace ("****************************");
			#endif

			file_ok = true;
			break;
		}
	}

	// BE SURE TO CLEANUP!!
	closedir(dp);			// close DIR stream 

	#ifdef TRACER_CODE
		Trace ("--> Dir Read Done");
	#endif

	// ONLY do the next step if file NOT found above
	int loop_count = 3;

	if (file_ok == false)
	{
		#ifdef TRACER_CODE
			Trace ("starting Quick Open Loop");
		#endif

		#ifdef TRACER_CODE
			Trace ("--> loop_count = " + int_to_string (loop_count));
		#endif

		for (int i = 1; i <= loop_count; i++)		
		{
			// dont forget the suffix!!!!
			if (file_exists_test (isam_dir_path + isam_file_name + ".vod") == false)
			{
				#ifdef TRACER_CODE
					Trace ("--> Quick File Open Failed - delay and try again");
				#endif

				if (i < loop_count)
					sleep(1);		// wait 1 second if more loops to do

				continue;			// another loop 
			}
			else	// open was ok - exit and return true!
			{
				#ifdef TRACER_CODE
					Trace ("****************************");
					Trace ("----> ISAM FILE EXISTS");
					Trace ("****************************");
				#endif

				return(true);			// exit back with true
			}
		}
	}

// gets here - return file_ok results
	#ifdef TRACER_CODE
		Trace ("--> does_isam_file_exist ENDED");
	#endif

	if (file_ok == true)
		return (true);
	else
		return (false);
}	

//**************************************************************
//** truncate string 
//**************************************************************
string ISAM::truncate_string (string string_in, unsigned int num_characters_to_keep)
{
	string results = string_in.substr(0, num_characters_to_keep);

	return(results);
}

//**************************************************************
//** REMOVE trailing slash from path name - IF any
//**************************************************************
string ISAM::remove_dir_trailing_slash (string file_path_name)
{
	// if path is / or ./ or ../, then the path is file as is
	if (strings_match (file_path_name, "/") || strings_match (file_path_name, "./") || strings_match (file_path_name, "../"))
		return(file_path_name);

	// else cleanup is needed
	unsigned int last_pos  = file_path_name.length() - 1;
	unsigned int start_pos = file_path_name.find_last_of("/");

	if (start_pos != string::npos && start_pos == last_pos)
		file_path_name = truncate_string (file_path_name, start_pos + 1);

	return(file_path_name);
}

//*******************************************************
//* see if a file exists
//*******************************************************
bool ISAM::file_exists_test (string file_path)
{
	#ifdef TRACER_CODE
		Trace ("file_exists_test for: " + file_path + trace_file_name ());
	#endif

	ifstream InputFile(file_path.c_str(), ios::in | ios::binary);

	if (!InputFile)				// open file failed
	{
		#ifdef TRACER_CODE
			Trace ("--> File MISSING");
		#endif

		InputFile.close();
		return (false);
	}
	else
	{
		#ifdef TRACER_CODE
			Trace ("--> File EXISTS");
		#endif

		InputFile.close();
		return (true);
	}
}
//*******************************************************
//* creates a VSAM file
//*******************************************************
bool ISAM::create_isam_file (void)
{
	#ifdef TRACER_CODE
		Trace ("create_isam_file called for: " + PATH_NAME + FILE_NAME);
	#endif


	string create_file_name = PATH_NAME + FILE_NAME;

	reset_error_codes();

	// vsam options using vsam TYPES!!
	WORD Options = 1;

	// short FAR PASCAL VsamCreate(						 // create a database
	//                      LPSTR      lpFilespec,       // database file name
	//                      LONG       GroupSize,        // group size
	//                      LPSTR      DbKey);           // database encryption key

	Options = 1;

	if (P_KEY_FIELD_NAME.length() == 0)
	{
		set_error (CREATE_FILE_INVALID, "Primary Key Name is blank. Primary Key Field Spec Missing?");
		return (false);
	}

	log_VSAM_call ("VsamCreate",
					create_file_name, 
					int_to_string (VSAM_GROUP_SIZE), 
					"ENCRYPTION_KEY");

	// according to Tony at VSAM, the primary key length here should aways be set to 252 (VSAM_MAX_KEY_LENGTH)
	// and to disregard what's in the manual. 
	short result = VsamCreate ((char *) create_file_name.c_str(), VSAM_GROUP_SIZE, (char *) ENCRYPTION_KEY.c_str());

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamCreate OKAY");
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamCreate Error: " + get_VSAM_error_msg (result));
		return (false);
	}
}

//*******************************************************
//* Gets / Returns VSAM Error Code Meaning
//*******************************************************
string ISAM::get_VSAM_error_msg (int error_num)
{
	string err_msg = int_to_string (error_num);
	string meaning = "";

	// determine VSAM error code meaning..
	switch (error_num)
	{
		case 0: 
			meaning = "VIS_OK - All is OK - Normal return";
			break;

		case 1:
			meaning = "VIS_NOT_FOUND - record not found.";
			break;

		case 2:
			meaning = "VIS_UPDATE_VIOLATION";
			break;

		case 3:
			meaning = "VIS_ACCESS_DENIED - Multi-user access";
			break;

		case 4:
			meaning = "VIS_BAD_DATASET_NUMBER - Invalid dataset ref - Re-Open Dataset";
			break;

		case 5:
			meaning = "VIS_BAD_FORMAT - Invalid format string - See Function Spec";
			break;

		case 6:
			meaning = "VIS_INVALID_KEY - Illegal char in key - See Function Spec";
			break;

		case 7:
			meaning = "VIS_BAD_PARAMETER_VALUE - Invalid value - See Function Spec";
			break;

		case 8:
			meaning = "VIS_BAD_FILE - File is not VSAM - See VsamOpen Spec";
			break;

		case 9:
			meaning = "VIS_ALREADY_EXISTS - Dataset Name exists	- See VsamCreate Spec";
			break;

		case 10:
			meaning = "VIS_NO_ROOM - No Note space left - See VsamWriteNote";
			break;

		case 11:
			meaning = "VIS_DISK_FULL - No Disk space left - Add Disk space";
			break;

		case 12:
			meaning = "VIS_OUT_OF_MEMORY - No Memory - Reduce Mem Use";
			break;

		case 13:
			meaning = "VIS_DOS_ERROR - DOS failure - See Function Spec";
			break;

		case 14:
			meaning = "VIS_DISK_ERROR - Disk read/write error - Fix Disk - then run VTOOL";
			break;

		case 15:
			meaning = "VIS_DATA_VALIDITY_CHECK - Corrupted dataset - Run VTOOL";
			break;

		case 16:
			meaning = "VIS_INVALID_SECONDARY_KEY - Secondary index too long or is NULL - See VsamPut Spec";
			break;

		case 17:
			meaning = "VIS_SEQUENCE_ERROR - Records are out of sort - Run DREMAKE or reatore dataset";
			break;

		case 18:
			meaning = "VIS_OUT_OF_FILE_HANDLES - No system file handles avail - Close some files";
			break;

		case 19:
			meaning = "VIS_BUSY - An operation was busy - Retry and notify the user.";
			break;

		case 20:
			meaning = "VIS_INVALID_LICENSE_KEY - Invalid License Key - See VsamOpen Spec";
			break;

		case 21:
			meaning = "VIS_FUNCTION_UNAVAILABLE - Function not implemented";
			break;

		case 22:
			meaning = "VIS_INTERRUPTED - Asynchronous operation interrupted the Program - See Function Spec ";
			break;

		case 101:
			meaning = "Bad Group header - Run VTOOL Rebuild or restore dataset from Backup";
			break;

		case 102:
			meaning = "Bad Group header - Run VTOOL Rebuild or restore dataset from Backup";
			break;

		case 103:
			meaning = "Bad Group header - Run VTOOL Rebuild or restore dataset from Backup";
			break;

		case 105:
			meaning = "A single partial record is larger than 1/2 group.";
			break;

		case 222:
			meaning = "Internal failure caused group split to create new group larger than maximum size of a group in this dataset";
			break;

		case 300:
			meaning = "Errors in the 300s can occur if some of the updates of a record have been completed but the system does not allow VSAM to extend the file to finish updating the record.";
			break;

		case 321:
			meaning = "Data record sub-parts are out of sequence";
			break;

		case 333:
			meaning = "Data record sub-parts are out of sequence";
			break;

		case 345:
			meaning = "Could not locate Primary record through an existing Secondary";
			break;

		case 346:
			meaning = "Secondary (Xref) couldn't be deleted - so we didn't delete primary";
			break;

		case 347:
			meaning = "Secondarys were deleted but could not delete the primary";
			break;

		case 348:
			meaning = "Secondary (Xref) update failed after successful primary update";
			break;

		case 444:
			meaning = "No address provided by Caller to return data from get -> Reboot?";
			break;

		case 501:
			meaning = "Record structure does not match the format string - Run VTOOL Rebuild or restore dataset";
			break;

		case 502:
			meaning = "Decode Buffer too small - Bad user defined type or Format string";
			break;

		case 503:
			meaning = "Decode Buffer too small - Bad user defined type or Format string";
			break;

		case 504:
			meaning = "A field in the Dataset record will not fit in User defined type";
			break;

		case 555:
			meaning = "Last partial piece of a record cannot be located - Record was truncated";
			break;

		case 666:
			meaning = "Dataset Record is too large for a VB String";
			break;

		case 765:
			meaning = "Raw read returned no data (NULL)";
			break;

		case 885:
			meaning = "Dataset activation failed, usually because network rights are not set properly, the Map file (.ISM) is corrupted or there is a problem creating or accessing the lock file. Run VTOOL Rebuild dataset";
			break;

		case 886:
			meaning = "Internal control block inconsistencies detected usually	Reboot a memory problem cause by a GPF or semiconductor failure";
			break;

		case 887:
			meaning = "Internal control block inconsistencies detected usually	Reboot a memory problem cause by a GPF or semiconductor failure";
			break;

		case 888:
			meaning = "Part of the VSAM dataset became unavailable to the software";
			break;

		case 1001:
			meaning = "A disk seek failure occurred in the data file - Fix the hardware. May need to restore dataset";
			break;

		case 1002:
			meaning = "A Data read/write failure occurred in the data file";
			break;

		case 1003:
			meaning = "The Group size found was too big. Run VTOOL Rebuild or restore dataset";
			break;

		case 1004:
			meaning = "A Disk seek failure occurred in the finder file. Fix the hardware. May need to restore the dataset";
			break;

		case 1005:
			meaning = "A finder read/write failure error occurred";
			break;

		case 1006:
			meaning = "A mismatch between the finder and actual data in. Rebuild the finder.";
			break;

		case 1014:
			meaning = "An error was detected during a group split process which generated a bad header.Run VTOOL Rebuild or restore dataset";
			break;

		case 1234:
			meaning = "Internal Consistence error in assembling the data record. Reboot. Usually a memory problem cause by a GPF or HW failure";
			break;

		case 2001:
			meaning = "Group Validity check - Record size was either too small too large for the Group. Run VTOOL or Restore dataset";
			break;

		case 2002:
			meaning = "Group Validity check - A record had no key";
			break;

		case 2003:
			meaning = "Group Validity check. Record size field exceeds Group limit";
			break;

		case 2005:
			meaning = "Group Validity check - The total of all record size fields does not match the size of the data in the group";
			break;

		default:
			meaning = "UNKNOWN VSAM Error Code";
			break;
	}

	// now add meaning string to results
	err_msg += " (" + meaning + ")";

	// return results
	return (err_msg);
}

//*******************************************************
//* sets the ISAM CLASS global error flags
//*******************************************************
void ISAM::set_error (unsigned int error_number, string error_message)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::set_error called for error_num: " + int_to_string(error_number) + trace_file_name ());
	#endif

	error   = true;
	err_num = error_number;
	err_msg = error_message;

	#ifdef TRACER_CODE
		Trace ("----> ERROR: " + err_msg);
	#endif

	return;
}

//*******************************************************
//* returns file_nmae string for Tracer dump
//*******************************************************
string ISAM::trace_file_name (void)
{
	return (" (" + FILE_NAME + ")");
}

//*******************************************************
//* unlocks the current file lock (if any)
//*******************************************************
bool ISAM::unlock (bool log_error = true)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::unlock called for key: " + VSAM_LOCK_KEY + trace_file_name ());
	#endif

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		if (log_error == true)
			set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");

		return (false);
	}

	// if nothing locked, then exit
	if (VSAM_LOCK_KEY.length() == 0)
	{
		#ifdef TRACER_CODE
			Trace ("--> No VSAM_LOCK_KEY so exit now.");
		#endif

		return (true);
	}

	// else an unlock is needed

	// short FAR PASCAL VsamUnlock (HANDLE DatasetNumber, LPSTR lpKey);

	reset_error_codes();

	log_VSAM_call ("VsamUnlock", "HANDLE: " + long_to_string (FILE_HANDLE), VSAM_LOCK_KEY);

	short result = VsamUnlock (FILE_HANDLE, (char *) VSAM_LOCK_KEY.c_str());

	if (result == VIS_OK)
	{
		VSAM_LOCK_KEY = "";			// be sure to clear this!!

		#ifdef TRACER_CODE
			Trace ("--> VsamUnlock OKAY.");
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamUnlock Error: " + get_VSAM_error_msg (result));
		return (false);
	}
}

//*******************************************************
//* locks the current record by p_key
//*******************************************************
bool ISAM::lock (string primary_key)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::lock called for p_key: " + primary_key + trace_file_name ());
	#endif

	// cant do this unless the file is opened
	if (FILE_OPEN != true)
	{
		set_error (FILE_NOT_OPEN, "Vsam Call Not Allowed - ISAM file is not open.");
		return (false);
	}

	// be sure to free the previous one, if it exists
	if (VSAM_LOCK_KEY.length () > 0)
	{
		#ifdef TRACER_CODE
			Trace ("--> need to free old file lock first.");
		#endif

		unlock ();
	}

	// if no primary key to lock, exit
	if (primary_key.length() == 0)
	{
		#ifdef TRACER_CODE
			Trace ("--> No p_key so exit without setting rec lock.");
		#endif

		return (false);
	}

	// else a lock is needed

	// short FAR PASCAL VsamLock (HANDLE DatasetNumber, LPSTR lpKey);

	// lock attempt loop
	for (int i = 1; i <= 3; i++)		// attempt lock 3 times 
	{
		reset_error_codes();

		log_VSAM_call ("VsamLock", "HANDLE: " + long_to_string (FILE_HANDLE), primary_key);

		short result = VsamLock (FILE_HANDLE, (char *) primary_key.c_str());

		if (result == VIS_OK)
		{
			VSAM_LOCK_KEY = primary_key;	// be sure to record this if OK

			#ifdef TRACER_CODE
				Trace ("--> VsamLock OKAY.");
			#endif

			return (true);
		}
		else
		{
			if (i == 3)		// after 3 seconds and 3 tries, fail this thing
			{
				set_error (result , "VsamLock Error: " + get_VSAM_error_msg (result));
				return (false);
			}
			else
			{
				#ifdef TRACER_CODE
					Trace ("--> VsamLock Failed - sleep 1 second and try LOCK again.");
				#endif

				sleep (1);		// sleep 1 second
				continue;
			}
		}
	}
}



//*******************************************************
//* Tracers a call to VSAM Routine
//*******************************************************
void ISAM::log_VSAM_call   (string function_name, 
							string P1 = "",
							string P2 = "",
							string P3 = "",
							string P4 = "",
							string P5 = "",
							string P6 = "",
							string P7 = "",
							string P8 = "")
{
	#ifdef TRACER_CODE
		Trace ("--> CALL " + function_name + " with the following parms:");
	#endif


	string msg = "--> " + function_name + " (";

	if (P1.length() != 0)
		msg += P1;

	if (P2.length() != 0)
		msg += ", " + P2;

	if (P3.length() != 0)
		msg += ", " + P3;

	if (P4.length() != 0)
		msg += ", " + P4;

	if (P5.length() != 0)
		msg += ", " + P5;

	if (P6.length() != 0)
		msg += ", " + P6;

	if (P7.length() != 0)
		msg += ", " + P7;

	if (P8.length() != 0)
		msg += ", " + P8;

	msg += ")";

	Trace (msg);

	return;
}

//*******************************************************
//* Object / Class Copy Constructor 
//*------------------------------------------------------
//*		housekeeping for copied pointers & dynamic arrays
//*------------------------------------------------------
//*		THIS IS ONLY CALLED when class is passed by value
//*     INSTEAD of by reference. If we place the copy
//*     constructor for this class inside the PRIVATE
//*     section of the class def, the class will HAVE to be
//*     passed by reference and not by value which is the
//*     best method! Really dont want to ever have to use
//*     this it can be a very tuff debug when problems
//*     occur. 
//*******************************************************
//ISAM::ISAM (const ISAM & CopiedObj)
//{
// this will never be called since we have moved the copy 
// constructor for this class to private section of class
// definition. ALWAYS PASS THIS OBJECT BY REFERENCE
// and NOT by value
//
//	#ifdef TRACER_CODE
//			Trace ("ISAM Class COPY CONSTRUCTOR Called.");
//  #endif
//
//	NUM_FIELDS_LOADED = CopiedObj.NUM_FIELDS_LOADED;
//	MAX_NUM_FIELDS    = CopiedObj.MAX_NUM_FIELDS;
//
//  ... work to be done to setup copied class
//
//	Trace ("ISAM Class COPY CONSTRUCTOR DONE.");
//
// THIS REPLACES the compiler's default copy object work so everything 
// in the class's constructor must be copied and initialized 
// mnually in here. This is not needed if we pass the object
// by reference and not by value. 
//}


//*******************************************************
//* APPENDS a new VSAM field definition to exisitng file
//* structure. This is a low level call directly to VSAM
//* and should not be confused with the class wrappers's
//* various ISAM::AddField routines. 
//*
//* NOTE: if this routine is called, VSAM will clear ALL
//*       record pointers to this file. VsamBOF, VsamEOF, 
//*       or VsamGET will have to be called to position
//*       the pointer in the DB. Therefore, this should
//*       be called at the start of FILE class init and 
//*       before any record work is done on DB so you
//*       dont lose your place in the DB.
//*******************************************************
bool ISAM::VSAM_AddFieldDef (string VSAM_field_name, string VSAM_field_type, bool is_index, long VSAM_dimension, 
							   long VSAM_width,      string VSAM_justify,    long& VSAM_field_num_assigned)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::VSAM_AddFieldDef for field_name: " + VSAM_field_name + trace_file_name ());
	#endif

	short result = 0;

//	short FAR PASCAL VsamAddField(                 // add a field to a database (new)
//                     HANDLE     DatasetNumber,    // handle of open database
//                     LPSTR      lpFieldName,      // name of field to add
//                     LPSTR      lpFieldType,      // field data type
//                     WORD       wDimension,       // NEW max instances of this field (dimension)
//                     BOOL       bIsIndex,         // true if field is to be an index field
//                     WORD       wWidth,           // field width
//                     LPSTR      lpJust,           // field justification
//                     LPWORD     lpFieldNumber);   // returned, number of added field

	string index_str = ""; 

	if (is_index == true)
		index_str = "INDEX";
	else
		index_str = "NOT INDEX";

	log_VSAM_call ("VSAM_AddFieldDef",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"\"" + VSAM_field_name + "\"",
					VSAM_field_type,
					index_str,
					"wDimension",
					long_to_string (VSAM_width),
					VSAM_justify,
					"&field_num_assigned");

	WORD field_num_assigned = 0;

	#ifdef TRACER_CODE
		Trace ("--> field_name being used: " + VSAM_field_name);
	#endif

	result = VsamAddField (FILE_HANDLE, (char *) VSAM_field_name.c_str(), (char *) VSAM_field_type.c_str(), (WORD) VSAM_dimension, is_index, (WORD) VSAM_width, (char *) VSAM_justify.c_str(), &field_num_assigned);

	VSAM_field_num_assigned = (long) field_num_assigned;

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamAddField OKAY. Field_num Assigned by VSAM: " + long_to_string (VSAM_field_num_assigned));
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamAddField Error: " + get_VSAM_error_msg (result));
		return (false);
	}

	return (true);
}


//*******************************************************
//* CHANGES a field's name/width/index attribute in an 
//* existing VSAM field definition. Should be called by 
//* field number. 
//*******************************************************
//----------------------------------------------------------------------------------------
// SPECIAL IMPLEMENTATION NOTES for changing "Find" Attribute:
//----------------------------------------------------------------------------------------
// If you wish to change this attribute's setting for an existing string field, 
// then ONLY certain changes are allowed depending on the current state of the index 
// in the file. They are as follows:
//
// T -> D  or  F -> P  or  P -> D  or  D -> P  
//
// Where:
//-----------------
// T = True     - all records in the file are currently indexed for this field (fully indexed)
// F = False    - no records in the file are indexed for this field
// P = Partial  - some of the records are indexed for this field. Only records added or 
//			 	  changed after setting the field's Find attribute to "P" will be indexed. 
//				  Run VsamRebuild to index all remaining records.
// D = Disabled - this field's previously active index (full or partial) has been disabled 
//                for all records.
//
// Therefore, you cannot perform T -> F or F -> T direct transformations on existing string fields. 
// The intermediate "P" and "D" codes must be used until a VsamRebuild can be run on the file.
//
// When a VsamRebuild is run on the file, the index field will be "cleaned up" for all of the file's 
// records, and the following index transformations will have been completed internally:
//
// P -> T  or  D -> F  
//
// Where:
//-----------------
// P = Was Partially indexed - only some records were indexed on this field.
// T = All records are now fully indexed for this field.
// D = Was Disabled index - (index existed but was not active)
// F = All Disabled indexes have been removed from the file
//----------------------------------------------------------------------------------------
bool ISAM::VSAM_ChangeFieldAttribute (string VSAM_field_num, string VSAM_attribute_name, string VSAM_attribute_value)
{
	#ifdef TRACER_CODE
		Trace ("ISAM::VSAM_ChangeFieldAttribute for: " + VSAM_field_num + trace_file_name ());
		Trace ("--> VSAM_attribute_name: "  + VSAM_attribute_name);
		Trace ("--> VSAM_attribute_value: " + VSAM_attribute_value);
	#endif

// short FAR PASCAL VsamSetFieldAttribute(        // set an attribute for a field (new)
//                      HANDLE     DatasetNumber,    // handle of open database
//                      LPSTR      lpField,          // field to access (field name or field number if %n)
//                      LPSTR      lpAttribueName,   // name of field attribue to store
//                      LPSTR      lpData);          // buffer containing attribuet data (0 term)
                                                     // NOTE: setting lpData to null "clears" the attribute
	log_VSAM_call ("VsamSetFieldAttribute",
					"HANDLE: " + long_to_string (FILE_HANDLE),
					"\"" + VSAM_field_num + "\"",
					"\""  + VSAM_attribute_name + "\"",
					"\""  + VSAM_attribute_value + "\"" );

	short result = VsamSetFieldAttribute (FILE_HANDLE, (char *) VSAM_field_num.c_str(), (char *) VSAM_attribute_name.c_str(), (char *) VSAM_attribute_value.c_str());

	if (result == VIS_OK)
	{
		#ifdef TRACER_CODE
			Trace ("--> VsamSetFieldAttribute OKAY.");
		#endif

		return (true);
	}
	else
	{
		set_error (result , "VsamSetFieldAttribute Error: " + get_VSAM_error_msg (result));
		return (false);
	}

	return (true);
}



