//************************************************************
// ISAM Class Definition
//*-----------------------------------------------------------
//** Written By:  Jay J Falconer
//**              Bitwise Software International, Inc.
//**              Phoenix, AZ 85028
//**              www.ShoppingQ.com
//************************************************************
// #include <stdlib.h>		// not needed
// #include <stdio.h>		// not needed
// #include <ctype.h>		// not needed

#include <fstream.h>		// required for file streams
#include <iostream.h>		// required for IO streams
#include <string>			// required for string classes
#include <sys/stat.h>		// required for directory reads
#include <sys/types.h>		// required for directory reads
#include <dirent.h>			// required for directory reads
#include <unistd.h>			// for sleep()
#include <vector.h>			// for dynamic arrays
#include <time.h>			// for get time func

#include "vsam.h"			// for VSAM
#include "linuxwin.h"		// for VSAM

// if you DO NOT want runtime tracer code compiled into your 
// object code, just just comment out the next 3 lines and recompile. 
#ifndef TRACER_CODE
  #define TRACER_CODE
#endif	

// Define ISAM class block
#ifndef ISAM_Class
#define ISAM_Class

// Start of ISAM low-level interface to VSAM
class ISAM
{
public:

//------------------------------------------------------------
// PUBLIC functions - callable from parent FILE class
//------------------------------------------------------------
	// CLASS CONSTRUCTOR / DESTRUCTOR
	ISAM::ISAM  (void);
	ISAM::~ISAM (void);

	// callable functions
	bool ISAM::ClassInit (string file_path_name, string file_name_str, unsigned int max_value = 100);
	bool ISAM::FileConfig (bool build_file_if_missing);

	bool ISAM::AddField        (string field_name, string field_type);
	bool ISAM::AddIndexField  (string field_name, string field_type, unsigned int max_key_length);
	bool ISAM::AddPrimaryKey  (string field_name, string field_type, unsigned int max_key_length);
	bool ISAM::AddCompoundKey (string index_name, string field_type, string field_names);

	bool ISAM::FileOpen        (void);
	bool ISAM::FileClose       (bool log_error = true);
	long ISAM::FileNumRecords  (void);

	bool ISAM::FreeVSAMBuffer  (void);

	bool ISAM::SetBOF (void);
	bool ISAM::SetEOF (void);

	void ISAM::Trace (string tracer_text, bool func_header = false);

	bool ISAM::AddRecord     (string primary_key, bool rec_must_not_exist = true);
	bool ISAM::UpdateRecord  (string primary_key, bool rec_must_exist     = false);
	bool ISAM::DeleteRecord  (string primary_key);

	bool ISAM::KillDataSet (void);

	// overloaded SetField value functions by data type to store
	bool ISAM::SetField   (string field_name, string& field_value);
	bool ISAM::SetField   (string field_name, int&    field_value);
	bool ISAM::SetField   (string field_name, long&   field_value);
	bool ISAM::SetField   (string field_name, float&  field_value);

	unsigned int ISAM::FileGetNumFields (void);

	bool ISAM::SetIndex      (string new_index_name);
	bool ISAM::GetNextRecord (bool load_data = true);
	bool ISAM::GetPrevRecord (bool load_data = true);

	bool ISAM::FindRecord (string index_str,   bool load_data = true);
	bool ISAM::GetRecord  (string primary_key, bool load_data = true);

	// overloaded GetField functions by the data type to return
	bool ISAM::GetField (string field_name, string& return_value);
	bool ISAM::GetField (string field_name, int&    return_value);
	bool ISAM::GetField (string field_name, long&   return_value);
	bool ISAM::GetField (string field_name, float&  return_value);

	string ISAM::Encrypt (string text_str, string cypher_key);
	string ISAM::Decrypt (string hex_str,  string cypher_key);

	string ISAM::GetFieldAttribute (string field_name,     string attrib_name);	

	bool   ISAM::GetFieldNameList (string &field_names,   long &field_count);
	bool   ISAM::GetFieldNumList  (string &field_numbers, long &field_count);

//------------------------------
// PUBLIC variables
//------------------------------
	bool rec_loaded;				// T or F

	bool         error;				// error control
	unsigned int err_num;			// error control
	string       err_msg;			// error control

	string PRIMARY_KEY;				// global string "PRIMARY"

//------------------------------
// PUBLIC constants
//------------------------------
	static const bool HEADER              = true;
	static const bool BUILD_IF_MISSING    = true;
	static const bool NO_BUILD_IF_MISSING = false;

//------------------------------------------------------------
// PRIVATE functions - not callable outside this class
//------------------------------------------------------------
private:

	// make COPY CONSTRUCTOR "PRIVATE" to force passing this OBJ by reference and not by value
	ISAM::ISAM (const ISAM & CopiedObj);

	// make ASSIGNMENT OPERATOR "PRIVATE" to stop assignment of this object to another 
	void operator= (const ISAM & CopiedObj);

	bool   ISAM::VSAM_AddFieldDef (string VSAM_field_name, string VSAM_field_type, bool is_index, long VSAM_dimension, long VSAM_width, string VSAM_justify, long& VSAM_field_num_assigned);
	bool   ISAM::VSAM_ChangeFieldAttribute (string VSAM_field_num, string VSAM_attribute_name, string VSAM_attribute_value);

	string ISAM::get_VSAM_error_msg (int error_num);

	bool   ISAM::check_new_field_parms     (string field_name, string field_type);
	string ISAM::remove_dir_trailing_slash (string path_name);

	string ISAM::truncate_string   (string string_in, unsigned int num_characters_to_keep);
	void   ISAM::reset_error_codes (void);

	unsigned int ISAM::parse_field (string data_field, 
									string result_array [], 
									string delimiter, 
									unsigned int max_fields);

	bool ISAM::does_isam_file_exist (string isam_dir_path, string isam_file_name);
	bool ISAM::file_exists_test     (string file_path);
	bool ISAM::create_isam_file     (void);

	string       ISAM::get_vsam_field_type (string bsi_field_type);
	unsigned int ISAM::get_field_num       (string field_name);
	unsigned int ISAM::get_index_field_num (string index_name);

	void ISAM::set_error (unsigned int error_number, string error_message);

	string ISAM::trace_file_name (void);

	bool ISAM::unlock (bool log_error = true);
	bool ISAM::lock   (string primary_key);

	void ISAM::log_VSAM_call   (string function_name, 
								string P1 = "",
								string P2 = "",
								string P3 = "",
								string P4 = "",
								string P5 = "",
								string P6 = "",
								string P7 = "",
								string P8 = "");

	void   ISAM::log_current_time (void);
	string ISAM::get_current_time (void);

	// special embedded generic functions (private)
	string ISAM::int_to_string   (const int& value_in);

	string ISAM::long_to_string (const HANDLE& handle_value_in);
	string ISAM::long_to_string (const long& value_in);

	string ISAM::float_to_string (const float& value_in);

	bool   ISAM::strings_match      (string str_1, string str_2);
	bool   ISAM::strings_dont_match (string str_1, string str_2);

	string ISAM::trim  (string string_in);
	string ISAM::ltrim (string string_in);
	string ISAM::rtrim (string string_in);
	string ISAM::ucase (string str_in);

	// for runtime code trace/debug (only works if TRACER_CODE defined) 
	void ISAM::check_debug_mode (void);
	void ISAM::clear_log_file   (void);

//------------------------------
// PRIVATE variables
//------------------------------
	// define dynamic arrays (vectors) that will be needed herein
	// the difference in total object size is 20K bigger using vectors
	// instead of dynamically allocated arrays from the free store.
	// Using vectors greatly reduces the possibility of memory leaks
	// so the little extra object code overhead is well worth it.
	vector<string> F_DATA;
	vector<string> F_NAME;
	vector<string> F_TYPE;
	vector<string> F_INDEX;
	vector<string> F_PARTS;
	vector<int>    F_LENGTH;
	vector<bool>   F_IN_C_KEY;

	unsigned int P_KEY_FIELD_LEN;
	string       P_KEY_FIELD_NAME;
	string		 P_KEY_FIELD_VALUE;

	string       CURR_INDEX_NAME;		// either "PRIMARY" for primary index, or one of the other index field names
	unsigned int CURR_INDEX_NUM;		// either 0 for primary index, or one of the other index fields
										// defined in the record format spec

	string PATH_NAME;					// Qfiles/ (be sure trailing slash included!!)
	string FILE_NAME;					// example Qsession (without the suffix and no path)

	unsigned int MAX_NUM_FIELDS;
	unsigned int NUM_FIELDS_LOADED;

	string VSAM_PASSWORD;
	string VSAM_LOCK_KEY;

	HANDLE FILE_HANDLE;				// VSAM flle handle

	bool FILE_OPEN;					// T or F

	string ENCRYPTION_KEY;

	// VSAM Record Buffer for LIB access
	GSTR FAR * VSAM_BUFFER;		// setup far * to structure of type GST

	//-----------------------------------------------------------
	// NOTE: this is used for retreiving a string field value
	//       from the VSAM_BUFFER. 
	// This controls the max string length per DB field
	// supported in the file system. 
	//
	// VSAM's Max string field size is 64K (65536).
	//
	// Recommended length = 15001 so not to waste too much
	// local process memory each time ISAM object is created
	// but still will allow a decent sized string field. 
	//-----------------------------------------------------------
	unsigned int MAX_STRING_LEN;
	char STRING_BUFFER [15001];			// 15,000 bytes

	// for runtime code/step debug
	bool TRACER_MODE;
	bool TRACER_MODE_CHECKED;
	bool TRACER_OFF;

	string TRACER_LOGFILE_NAME;
	string DEBUG_FILE_NAME;

	// ISAM class error equates
	enum ERROR_EQUATES			// list of error numbers - assignment number dont matter - just need to be unique
	{
		FILE_NAME_INVALID,
		FILE_NAME_MISSING,
		TOO_MANY_FIELDS,
		P_KEY_MISSING,
		FIELD_NAME_INVALID,
		FIELD_TYPE_INVALID,

		FILE_OPEN_ERROR,
		FILE_CLOSE_ERROR,

		P_KEY_FIELD_ERROR,
		INDEX_FIELD_ERROR,
		C_KEY_FIELD_ERROR,

		ADD_RECORD_FAILED,
		DELETE_RECORD_FAILED,
		UPDATE_RECORD_FAILED,

		CREATE_FILE_INVALID,
		INDEX_NAME_INVALID,

		FILE_NOT_OPEN,
		GET_NEXT_REC_ERROR,
		GET_PREV_REC_ERROR,

		KILL_DATASET_ERROR,

		FILE_MUST_NOT_BE_OPEN,

		END_LIST
	};

//------------------------------
// PRIVATE CONSTANTS
//------------------------------
	// VSAM CONSTANTS / LIMIT PARAMETERS 
	static const unsigned int VSAM_GROUP_SIZE          = 32768;		// this allows largest file size possible
	static const unsigned int VSAM_MAX_KEY_LENGTH      = 252;		// the largest possible allowed by VSAM
	static const unsigned int VSAM_KEY_FIELD_NUM_LIMIT = 157;		// the largest possible allowed by VSAM

	static const bool NO_LOG_ERROR = false;

	long LOG_TIME_COUNT;
};

#endif	// end ISAM_Class
