//********************************************************************
//** Qdriver.cgi
//********************************************************************
//** ISAM Class Tutorial
//**------------------------------------------------------------------
//** Written By:  Jay J Falconer
//**              Bitwise Software International, Inc.
//**              Phoenix, AZ 85028
//**              www.ShoppingQ.com
//*-----------------------------------------------------------
//* NO WARRANTY:
//*-----------------------------------------------------------
//* This code is provided as part of VSAM for Linux
//* It is provided "AS IS" and the Author does not offer support 
//* nor any warranties or guarantees. 
//* USE THIS CODE at your own risk. 
//*******************************************************

#include "Quser.h"

string err_msg      = "";
string vsam_err_msg = "";

//*-----------------------------------------------------------
//** STEP 1: 
//*-----------------------------------------------------------
//** In the ISAM class (low level interface to VSAM) source
//** code (ISAM.cpp) you need to change the following global
//** variables:
//**
//**	VSAM_PASSWORD
//**	ENCRYPTION_KEY
//** 
//** VSAM_PASSWORD - needs to contain your assigned VSAM password  
//** as assigned by Software Source that activates your copy of VSAM. 
//** 
//** ENCRYPTION KEY - can be any string of characters. This is used
//** as the cypher key to encrypt your VSAM records. Once you assign
//** and use this key for a file, NEVER change it. It must remain the
//** same in order for VSAM to decrypt the records and read them. 
//** 
//** These variables are contained in the ISAM::ISAM constructor 
//** function. Please change them now. 
//** 
//** Other than that, the ISAM, class should never have to be 
//** changed (unless you want to turn the tracer code compilation 
//** OFF. Please see the notes in the ISAM class header file regarding
//** this runtime feature. See the "#ifndef TRACER_CODE" directive block. 
//*-----------------------------------------------------------

//*-----------------------------------------------------------
//** STEP 2: 
//*-----------------------------------------------------------
//** Next, you need a file class like the sample "Quser.cpp" 
//** class that is included. This is a parent class to the
//** ISAM class that actually defines a database file. This
//** file class works with the ISAM class to costruct and
//** maintain a VSAM database file used in your application. 
//** 
//** In the Quser.cpp example, this is defining a DB file
//** that will hold User name and password type information
//** for an application. Read through the source code. Pay
//** close attention to how a file's record structure is 
//** defined in the "bool Quser::define_fields (void)" function.
//** This section tells ISAM what the records look like for this
//** file. ISAM takes over from there automatically. 
//** 
//** The actual fields for this file's records are defined in 
//** this class as well - see the "PUBLIC variables" section
//** of the Quser.h file. 
//** 
//** Any changes made to the fields for this file, must be 
//** reflected in the header file and in the following 
//** Quser.cpp functions:
//** 
//**	bool Quser::define_fields (void)
//**	bool Quser::clear_fields (void)
//**	bool Quser::set_all_fields (string primary_key)
//**	bool Quser::get_all_fields (bool load_all_fields = true)
//** 
//** There are certain restictions to the types of record changes
//** that can be made to a file once the file has been created. 
//** Please review the notes in the "bool Quser::define_fields (void)"
//** function. 
//** 
//** For each file in your application, simply copy the Quser.cpp
//** and Quser.h files, rename them to reflect the new file's
//** function (like Qinventory.cpp and Qinventory.h). Then change
//** their source so all "Quser" references are now "Qinventory".
//** That creates a new file class called Qinventory. Next, change
//** the field specs (record structure) in the PUBLIC variables
//** section of the Qinventory.h file and the 4 field spec functions
//** listed above inthe Qinventory.cpp file. 
//** 
//** The rest of the functions in Qinventory.cpp need never be 
//** changed but must be included in order to access the ISAM
//** class functions. Just be sure they are all copied to each
//** new file class you create. 
//*-----------------------------------------------------------

//*-----------------------------------------------------------
//* NO WARRANTY:
//*-----------------------------------------------------------
//* This code is provided as part of VSAM for Linux
//* It is provided "AS IS" and the Author does not offer support 
//* nor any warranties or guarantees. 
//* USE THIS CODE at your own risk. 
//*-----------------------------------------------------------

//*-----------------------------------------------------------
//** Using the Quser.cpp class:
//*-----------------------------------------------------------
//** In the code below, you will see many file open calls
//** but NO file close calls! That is because the ISAM
//** class (referenced by the Quser file class) takes care
//** of this automatically whenever the program exits
//** or the Quser object goes out of socope. If you want
//** to call it manually, you can by simply using
//** Quser_DB.FileClose(); below. 
//*-----------------------------------------------------------

//*-----------------------------------------------------------
//** Declare and Open the Database
//*-----------------------------------------------------------
//** TIP: in the Quser class implementation, if the file
//** does not exist, it is created automatically. Please
//** see the code in the Quser::Quser constructor func.
//*-----------------------------------------------------------
void sample_1 (void)
{
	Quser user_DB;					// declare user database object
									// if file missing, it is created.
	if (user_DB.open() == false)	// now try to open it
	{
		err_msg = "Quser file could not be opened.";
		return;
	}
}

//********************************************************
//** Fetch a record by PRIMARY KEY
//********************************************************
void sample_2 (void)
{
	string rec_num = "1234567";		// primary key rec_num

	Quser user_DB;					// declare user database object
									// if file missing, it is created.

	if (user_DB.open() == false)	// now try to open it
	{
		err_msg = "Quser file could not be opened.";
		return;
	}

	user_DB.set_index ("PRIMARY");	// set usage to primary key index (redundant - see below)

	if (user_DB.get_rec (rec_num) == false)		// NOTE: get_rec is ALWAYS by PRIMARY index
	{											// so the set_index call is really not needed 
												// before this function call.
		if (user_DB.error() == true)
			vsam_err_msg = user_DB.get_error_msg();

		err_msg = "The record was not found.";
		return;
	}

// NOTE: if the get_rec function call == true, the Quser data fields
//       will be automatically loaded for use. 
	
}

//********************************************************
//** Find a record by a secondary index or compound index
//** and compare its password to the one entered
//********************************************************
void sample_3 (void)
{
	string frm_email_addr = "sales@shoppingq.com";
	string frm_password   = "testing";

	Quser user_DB;					// declare user database object
									// if file missing, it is created.

	if (user_DB.open() == false)	// now try to open it
	{
		err_msg = "Quser file could not be opened.";
		return;
	}

	user_DB.set_index ("email_addr");	// set the current index to secondary index named "email_addr"
										// this is required in order for the find_rec to be run against
										// the correct index

	if (user_DB.find_rec (frm_email_addr) == false)		// DB operation failed?
	{
		if (user_DB.error() == true)					// did VSAM report an error?
			vsam_err_msg = user_DB.get_error_msg();		// if so grab it

		err_msg = "The Email Address was not found.";
		return;
	}

// NOTE: if the find_rec function call == true, the Quser data fields
//       will be automatically loaded for use below, 
	if (frm_password != user_DB.password)
	{
		err_msg = "The Password you entered did not match the password we have in our records.";
		return;
	}
}

//********************************************************
//** ADDs a new record (by primary key)
//********************************************************
void sample_4 (void)
{
	Quser user_DB;					// declare user database object
									// if file missing, it is created.

	if (user_DB.open() == false)	// now try to open it
	{
		err_msg = "Quser file could not be opened.";
		return;
	}

	user_DB.clear_rec();			// clear ALL record fields for add

// load Quser variable fields before add
	user_DB.rec_num         = "987654321";			// is a logical field - this is not really needed

	user_DB.email_addr      = "jay@shoppingq.com";
	user_DB.user_name       = "Jay Falconer";
	user_DB.last_access     = "";
	user_DB.web_site_addr   = "http://www.shoppingq.com";
	user_DB.password        = "testing2";
	user_DB.co_name         = "Bitwise Software International, Inc.";
	user_DB.co_type         = "Software Developer";
	user_DB.addr            = "10034 N. 25th Street";
	user_DB.city            = "Phoenix";
	user_DB.state           = "AZ";
	user_DB.zip             = "85028";
	user_DB.country         = "USA";
	user_DB.phone           = "602-555-1111";
	user_DB.web_host_name   = "www.shoppingQ.com";
	user_DB.web_server_type = "Linux";
	user_DB.store_addr      = "http://www.webstoreaddress.com";
	user_DB.store_version   = "1.0";
	user_DB.store_user_id   = "jayfalconer";
	user_DB.store_password  = "testing2";

	if (user_DB.add_rec (user_DB.rec_num) == false)
	{
		if (user_DB.error() == true)					// did VSAM report an error?
			vsam_err_msg = user_DB.get_error_msg();		// if so grab it

		err_msg =  "Quser record add failed.";
		return;
	}

}

//********************************************************
//** Updates a new record after a get (by primary key)
//********************************************************
void sample_5 (void)
{
// Let's assume user is changing their email address and password
	string frm_email_addr = "sales@shoppingq.com";
	string frm_password   = "testing";

// this is the primary rec key to update
	string rec_num = "1234567";		// primary key rec_num to update

	Quser user_DB;					// declare user database object
									// if file missing, it is created.

	if (user_DB.open() == false)	// now try to open it
	{
		err_msg = "Quser file could not be opened.";
		return;
	}

// first, let's get the record by its primary key
	if (user_DB.get_rec (rec_num) == false)		// NOTE: get_rec is ALWAYS by PRIMARY index
	{											// so a set_index call is not needed 
												// before this function call.
		if (user_DB.error() == true)
			vsam_err_msg = user_DB.get_error_msg();

		err_msg = "The record was not found.";
		return;
	}

// NOTE: if the get_rec function call == true, the Quser data fields
//       will be automatically loaded for use below. 

// now let's change ONLY the fields we need updated
	user_DB.email_addr      = frm_email_addr;
	user_DB.password        = frm_password;

// now put record back:
//		NOTE: record locking is automatic in the ISAM class which
//		is called by the Quser class. 
	if (user_DB.update_rec (rec_num) == false)
	{
		if (user_DB.error() == true)					// did VSAM report an error?
			vsam_err_msg = user_DB.get_error_msg();		// if so grab it

		err_msg =  "Quser record update failed.";
		return;
	}

}

//********************************************************
//** Fetch information about the database file from VSAM
//********************************************************
void sample_6 (void)
{
	Quser user_DB;					// declare user database object
									// if file missing, it is created.

	if (user_DB.open() == false)	// now try to open it
	{
		err_msg = "Quser file could not be opened.";
		return;
	}

// declare variables for results
	unsigned int number_fields = 0;
	long number_records        = 0;
	string vsam_record_format  = "";

// now fetch info for this file
	number_fields       = user_DB.get_num_fields ();
	number_records      = user_DB.get_num_recs   ();

// this shows how to retrieve a list of field names for this file
	string field_list = "";
	user_DB.get_field_name_list (field_list);

// this shows how to retrieve a list of field numbers for this file
	string num_list = "";
	user_DB.get_field_num_list (num_list);

// this shows how to encrypt a string and decrypt a string
// -----------------------------------------------------------
// NOTE: the ISAM class does this automatically for the entire
//       database file and its records. The code below is just
//       showing you how to use VSAM's encryption routines
//       manually in case you need to encrypt non-file strings 
//       for some reason (like encrypting browser cookies). 
// -----------------------------------------------------------

// Defines a cypher key string to be used to "seed" the VSAM encryption functions. 
	string cypher_key = "Cypher encryption key - can be any string of characters";

// Encrypts a text_string using the cypher_key - returns encrypted hex_string result 
	string encrypted_hex_string = user_DB.encrypt ("Text String to Encrypt Here", cypher_key);

// Decrypts the encrypted hex_string using the same cypher_key - returns original text string 
	string decrypted_hex_string = user_DB.decrypt (encrypted_hex_string, cypher_key);

	return;
}

//********************************************************
//** Loop to read thru all records by user name
//** to manually count the records and store the last 
//** user_name read
//********************************************************
void sample_7 (void)
{
	Quser user_DB;					// declare user database object
									// if file missing, it is created.

	if (user_DB.open() == false)	// now try to open it
	{
		err_msg = "Quser file could not be opened.";
		return;
	}

	user_DB.set_index ("user_name");	// set index to use
										// this is required in order for the find_rec to be run against
										// the correct index
										// REMEMBER - this also auto sets pointer
										// to start of file (BOF) by ths index!

	long rec_count = 0;

	string last_user_name = "";				// variable to hold last user name read

// NOTE: when the get_next_rec function call == true, the Quser data fields
//       will be automatically loaded each time thru the loop. 

// read records from start of file by user_name index
	while (true)
	{
		if (user_DB.get_next_rec () == false)	// no next record?
			break;								// break out of loop

		rec_count++;							// count records
		last_user_name = user_DB.user_name;		// load record field for processing...
	}

}

//********************************************************
//** Loop to read thru records by user name
//** looking for names that start with "sm"
//** uses lookup to start in the file just before
//** names starting with "sm"
//********************************************************
void sample_8 (void)
{
	Quser user_DB;					// declare user database object
									// if file missing, it is created.

	if (user_DB.open() == false)	// now try to open it
	{
		err_msg = "Quser file could not be opened.";
		return;
	}

	user_DB.set_index ("user_name");	// set index to use
										// this is required in order for the find_rec to be run against
										// the correct index
										// REMEMBER - this also auto sets pointer
										// to start of file (BOF) by ths index!

	bool pointer_set = false;

	string lookup_string = "sm";

// NOTE: when the find_rec or get_next_rec functions below == true, the Quser data fields
//       will be automatically loaded each time thru the loop. 
	while (true)
	{
		if (pointer_set == false)	// if first time thru loop, we need to set file pointer
		{
			// let's try to position file pointer to the very first record
			// that contains exactly "sm" as the user name. If it's there,
			// the record fields will be loaded. If not there, it gets the 
			// very first record after the file pointer is set
			if (user_DB.find_rec (lookup_string) == false)	
			{
				if (user_DB.get_next_rec () == false)		// else just get next record after "sm"
					break;									// exit read loop if EOF
			}
			// flag that we have done this one time
			pointer_set = true;
		}
		else		// else, not first time thru loop, so we need to just get the next record
		{
			if (user_DB.get_next_rec () == false)				// NO next record in the file?
				break;											// exit read loop if EOF
		}

		// if gets here, we have a first record or a next record from above
		// so now see if and where "sm" occurs in the user_name field
		unsigned int match_pos = user_DB.user_name.find (lookup_string); 

		if (match_pos != 0)		// if "sm" not in pos 0 of field
		{
			break;				// done reading records that start with "sm" - exit read loop
		}
		else
		{
			// else process this record and continue reading records
			// do something here...
			continue;
		}
	}
}

