//************************************************************
//* Quser File Class 
//*-----------------------------------------------------------
//*	USER FILE Interface Class for VSAM Class Wrapper
//*-----------------------------------------------------------
//*	NOTES:	This file class could be coded to be "derived"
//*			from the ISAM class instead of containing
//*			the ISAM class. However, using "composition"
//*			instead of "derivation" provides a better level
//*			of program readability among the classes since
//*			all application calls must come into and thru
//*			this intermediate class. 
//*-----------------------------------------------------------
//* Written By:  Jay J Falconer
//*              Bitwise Software International, Inc.
//*              Phoenix, AZ 85028
//*              www.ShoppingQ.com
//*-----------------------------------------------------------
//* TRACER NOTES:
//*-----------------------------------------------------------
//* The DB.Trace() functions here provide the ability to debug
//* the code at runtime. Step logging will be done if the
//* class code finds the file named in DEBUG_FILE_NAME in the same directory
//* as the compiled script program. If this file is missing,
//* the code will not log its steps and calls. Since it's
//* very difficult to perform debugging in a stateless web
//* environment when a script program only runs for a 
//* moment, this feature is important to help diagnose
//* problems during end-user runtime. Logging will be 
//* done to a text file called "logfile.txt". Since
//* the logging will slow down the code, it's recommended
//* that the debug,txt file NOT exist unless you need
//* to locate a problem during development or runtime. 
//* While the extra tracer code does make the object code
//* much bigger, the tradeoff in being able to remotely
//* debug the usage of this code at runtime is well worth it. 
//*-----------------------------------------------------------
//* CODING STYLE:
//*-----------------------------------------------------------
//* While there are many ways to code the same task and there
//* are certainly shorter ways to accomplish the tasks below, 
//* the coding methods used here are to make the code 
//* as readable and maintainable as possible. It is presented
//* as both a useable production tool as well as a VSAM C++ 
//* implementation tutotial.
//*-----------------------------------------------------------
//* NO WARRANTY:
//*-----------------------------------------------------------
//* This code is provided as part of VSAM for Linux
//* It is provided "AS IS" and the Author does not offer support 
//* nor any warranties or guarantees. 
//* USE THIS CODE at your own risk. 
//************************************************************

#include "Quser.h"

//************************************************************
//* Quser FILE Class Constructor
//*-----------------------------------------------------------
//*	This is called after the DB.ISAM constructor
//************************************************************
Quser::Quser (void)
{
	#ifdef TRACER_CODE
		DB.Trace ("Quser Class Constructor Called.");
	#endif


// UNIQUE VARIABLES for this DB FILE
	string path_name			= "Qfiles/";	// trailing slash! use blank or "./" for curr dir
	string file_name			= "Quser";		// NO file extension

	unsigned int max_num_fields = 75;			// max number for fields - for efficient vector pre-size

// THIS MUST BE CALLED before anything else can be done with this file!
	DB.ClassInit (path_name, file_name, max_num_fields);

	// tell VSAM what order and type the fields and indexes are for this file.
	define_fields();

	// clear all Quser class field variables
	clear_fields();

	// now take all field data above and create VSAM required format string
	// --> default index will be p_key
	DB.FileConfig (DB.BUILD_IF_MISSING);

	// this is redundant but shows how to set key to primary key
	DB.SetIndex (DB.PRIMARY_KEY);

}

//*******************************************************
//* NOTES:
//*------------------------------------------------------
//*		Only the next few functions need to be changed
//*		to control how the File class's fields are 
//*     defined, cleared, added, and updated. 
//*
//*     Dont forget to define the field variables in
//*     the public variables section of this class's 
//*     header file. 
//*******************************************************

//*******************************************************
//* Defines the Quser Record Fields
//*******************************************************
bool Quser::define_fields (void)
{
	#ifdef TRACER_CODE
		DB.Trace ("Quser::define_fields called");
	#endif

//------------------------------------------------------------------------------
// IMPORTANT: 
//------------------------------------------------------------------------------
//		--> Be sure "max_num_fields" value is enough to hold all these fields!!
//------------------------------------------------------------------------------

//-----------------------------------------------------------------------------------------------
// Load the PRIMARY KEY field (MUST BE STRING, must be defined, and can only one be 1 per file) 
//       
// PRIMARY KEY field definition MUST ALWAYS be the first field defined in the class!
// Once the primary key has been defined in the file, it cannot be changed (e.g. its length). 
//-----------------------------------------------------------------------------------------------
	DB.AddPrimaryKey ("rec_num", "S", 30);		// same as primary key (time+IP+seq)

//-----------------------------------------------------------------------------------------------
// NOTES:the remaining fields can be any type of field except PRIMARY KEY.
//-----------------------------------------------------------------------------------------------
//		 Once the VSAM file has been created, you can ONLY change the field name, width, and 
//       index flag for a field - NOT ANY of the remaining attributes: field number, delete flag, 
//       data type, & class cannot be changed.  
//
//       Once the VSAM file has been created, you must NOT change the order of the fields below! 
//	     also, do not use the VsamDelete function to delete a VSAM field definition or else the 
//       class's field order assumptions will not work. 
//       
//       Once the VSAM file has been created, you must NOT remove any of the fields below or 
//       else the class's field order assumptions will not work. 
//       
//		 Once the VSAM file has been created, to change a field's name, width, index flag, or 
//       justify type, simply change it's specs in the list below. This class will take care of 
//       the rest automatically. 
//       
//       Once the VSAM file has been created... if you need to change a field from one data type 
//       to another (string to long for example), then you need to change the name of original field 
//       to a "dummy" or unused name and then ADD the new field type (with the original field name) 
//       to the END OF THE field list (as a brand new field). The data in the old field will NOT be 
//       copied to the new field - it will however still exist in the file but will only be accessable 
//       under the fields "dummy" or unused field name you assigned. 
//
//	     Once the VSAM file has been created, DO NOT use the VsamDelete function to delete a field 
//  	 definition or else the class's field order assumptions will not work!
//
//       Once the VSAM file has been created,.. If you need to ADD brand new fields to the file structure, 
//    	 you must add them at the THE VERY END OF the field list below. This class will automatically 
//  	 inform VSAM of the appended field. VSAM will then internally change the data format in the 
//	     database when the class is loaded and initialized. NEW FIELDS MUST BE ADDED TO THE END OF THE 
//       LIST!
	
//       Therefore, once a record format has been defined below and the VSAM file is created, 
//       you CANNOT make ANY changes other than to the name of the field, its width and its index 
//	     flag. Everything else must remain constant. 
//-----------------------------------------------------------------------------------------------

// Load single field alt/secondary indexes (MUST BE STRINGS)
	DB.AddIndexField ("email_addr",    "S", 90);		// 90 chars + p_key added by VSAM
	DB.AddIndexField ("user_name",     "S", 60);		// 60 chars + p_key added by VSAM
	DB.AddIndexField ("last_access",   "S", 20);		// time in seconds + p_key added by VSAM
	DB.AddIndexField ("web_site_addr", "S", 90);		// 90 chars + p_key added by VSAM

// load normal data fields
	DB.AddField ("password", "S");
	DB.AddField ("co_name", "S");
	DB.AddField ("co_type", "S");
	DB.AddField ("addr", "S");
	DB.AddField ("city", "S");
	DB.AddField ("state", "S");
	DB.AddField ("zip", "S");
	DB.AddField ("country", "S");
	DB.AddField ("daytime_phone", "S");
	DB.AddField ("web_host_name", "S");
	DB.AddField ("web_server_type", "S");
	DB.AddField ("store_addr", "S");
	DB.AddField ("store_version", "S");
	DB.AddIndexField ("store_user_id", "S", 30);
	DB.AddField ("store_password", "S");

	DB.AddField ("night_phone", "S");

// new field to add to database after the fact

// Example of how to define logical compound STRING key fields: 
//	DB.AddCompoundKey ("name_index", "S", "user_name+email_addr");

// If you need to ADD brand new fields to the file structure, you must add them at the 
// THE END OF the field list below. This class will automatically inform VSAM of the appended 
// field. VSAM will then internally change the data format in the database when the class is 
// loaded and initialized. 

	return (true);
}

//*******************************************************
//* Clears the Quser Record Fields
//*******************************************************
bool Quser::clear_fields (void)
{
	#ifdef TRACER_CODE
		DB.Trace ("Quser::clear_fields called");
	#endif

	rec_num         = "";
	email_addr      = "";
	user_name       = "";
	last_access     = "";
	web_site_addr   = "";
	password        = "";
	co_name         = "";
	co_type         = "";
	addr            = "";
	city            = "";
	state           = "";
	zip             = "";
	country         = "";
	phone           = "";
	web_host_name   = "";
	web_server_type = "";
	store_addr      = "";
	store_version   = "";
	store_user_id   = "";
	store_password  = "";
	night_phone     = "";

	return (true);
}

//*******************************************************
//* Loads class field values into curr VSAM field buffer
//*******************************************************
bool Quser::set_all_fields (string primary_key)
{
	#ifdef TRACER_CODE
		DB.Trace ("Quser::set_all_fields called for p_key: " + primary_key);
	#endif

	DB.SetField ("rec_num",         rec_num);		// primary key - this statment not really needed!!
	DB.SetField ("email_addr",      email_addr);
	DB.SetField ("user_name",       user_name);
	DB.SetField ("last_access",     last_access);
	DB.SetField ("web_site_addr",   web_site_addr);
	DB.SetField ("password",        password);
	DB.SetField ("co_name",         co_name);
	DB.SetField ("co_type",         co_type);
	DB.SetField ("addr",            addr);
	DB.SetField ("city",            city);
	DB.SetField ("state",           state);
	DB.SetField ("zip",             zip);
	DB.SetField ("country",         country);
	DB.SetField ("phone",           phone);
	DB.SetField ("web_host_name",   web_host_name);
	DB.SetField ("web_server_type", web_server_type);
	DB.SetField ("store_addr",      store_addr);
	DB.SetField ("store_version",   store_version);
	DB.SetField ("store_user_id",   store_user_id);
	DB.SetField ("store_password",  store_password);

	DB.SetField ("night_phone",     night_phone);
	
	return (true);
}

//*******************************************************
//* Loads a record's fields into class variables
//*------------------------------------------------------
//*		assumes VSAM_BUFFER has been loaded ok by
//*		last call to ISAM
//*******************************************************
bool Quser::get_all_fields (bool load_all_fields = true)
{
	#ifdef TRACER_CODE
		DB.Trace ("Quser::get_all_fields called");
	#endif

	// always load the primary record key regardless
	#ifdef TRACER_CODE
		DB.Trace ("--> load primary key");
	#endif

	DB.GetField (DB.PRIMARY_KEY, rec_num);

	// Exit if all we want is primary key
	if (load_all_fields != true)
		return (true);

	// now load all the remaining data fields...
	#ifdef TRACER_CODE
		DB.Trace ("--> load all remaining data fields too");
	#endif

	DB.GetField ("email_addr",      email_addr);
	DB.GetField ("user_name",       user_name);
	DB.GetField ("last_access",     last_access);
	DB.GetField ("web_site_addr",   web_site_addr);
	DB.GetField ("password",        password);
	DB.GetField ("co_name",         co_name);
	DB.GetField ("co_type",         co_type);
	DB.GetField ("addr",            addr);
	DB.GetField ("city",            city);
	DB.GetField ("state",           state);
	DB.GetField ("zip",             zip);
	DB.GetField ("country",         country);
	DB.GetField ("phone",           phone);
	DB.GetField ("web_host_name",   web_host_name);
	DB.GetField ("web_server_type", web_server_type);
	DB.GetField ("store_addr",      store_addr);
	DB.GetField ("store_version",   store_version);
	DB.GetField ("store_user_id",   store_user_id);
	DB.GetField ("store_password",  store_password);

	DB.GetField ("night_phone",     night_phone);

	return (true);
}

//*******************************************************
//* NOTE:
//*------------------------------------------------------
//*		Everything below this point is simply an
//*		interface to the low level ISAM class 
//*     functions. Nothing needs to be changed below.
//*
//*		These wrapper routines need to be
//*     copied into each File class created to 
//*     use the ISAM class. 
//*******************************************************

//*******************************************************
//* Gets Field Attributes for request field in DB
//*******************************************************
bool Quser::get_field_attributes (string field_name)
{
	#ifdef TRACER_CODE
		DB.Trace ("Quser::get_field_attributes for field: " + field_name + " called.");
	#endif

	ATTRIBUTES.OK      = false;
	ATTRIBUTES.Fnumber = "";
	ATTRIBUTES.Fclass  = "";
	ATTRIBUTES.Ftype   = "";
	ATTRIBUTES.Findex  = "";
	ATTRIBUTES.Fwidth  = "";
	ATTRIBUTES.Fjust   = "";
	ATTRIBUTES.Fdelete = "";

	ATTRIBUTES.Fname = DB.GetFieldAttribute (field_name, "Fnam");

	if (ATTRIBUTES.Fname.length() == 0)
		return (false);

	ATTRIBUTES.OK = true;

	ATTRIBUTES.Fnumber = DB.GetFieldAttribute (field_name, "Fnum");
	ATTRIBUTES.Fclass  = DB.GetFieldAttribute (field_name, "Fcls");
	ATTRIBUTES.Ftype   = DB.GetFieldAttribute (field_name, "Ftyp");
	ATTRIBUTES.Findex  = DB.GetFieldAttribute (field_name, "Find");
	ATTRIBUTES.Fwidth  = DB.GetFieldAttribute (field_name, "Fwid");
	ATTRIBUTES.Fjust   = DB.GetFieldAttribute (field_name, "Fjst");
	ATTRIBUTES.Fdelete = DB.GetFieldAttribute (field_name, "Fdel");

	return (true);
}

//*******************************************************
//* Adds a record / assumes all the fields are loaded
//*******************************************************
bool Quser::add_rec (string primary_key, bool rec_must_not_exist = true)
{
	#ifdef TRACER_CODE
		DB.Trace ("Quser::add_rec called for primary_key: " + primary_key);
	#endif

	DB.FreeVSAMBuffer();				// clear record buffer before ADDs
									// this forces VSAM to create an all new record buffer
									// with empty values for all fields defined in the spec

	set_all_fields (primary_key);	// load field values for this FILE class

	return (DB.AddRecord (primary_key, rec_must_not_exist));
}

//*******************************************************
//* update record / assumes all the fields are loaded
//*******************************************************
bool Quser::update_rec (string primary_key, bool rec_must_exist = false)
{
	#ifdef TRACER_CODE
		DB.Trace ("Quser::update_rec called for primary_key: " + primary_key);
	#endif

	DB.FreeVSAMBuffer();				// clear record buffer before UUPDATE
									// this forces VSAM to create an all new record buffer
									// with empty values for all fields defined in the spec

	set_all_fields (primary_key);	// load field values for this FILE class

	return (DB.UpdateRecord (primary_key, rec_must_exist));
}

//*******************************************************
//* gets a record by PRIMARY KEY
//*******************************************************
bool Quser::get_rec (string primary_key, bool load_data = true)
{
	#ifdef TRACER_CODE
		DB.Trace ("Quser::get_rec called for primary_key: " + primary_key);
	#endif

	clear_rec(load_data);			// clear VSAM buffer and all Quser fields

	if (DB.GetRecord (primary_key, load_data) == true)
	{
		#ifdef TRACER_CODE
			DB.Trace ("Quser::get_rec OKAY. Now load fields.");
		#endif

		get_all_fields (load_data);
		return (true);
	}
	else
	{
		#ifdef TRACER_CODE
			DB.Trace ("Quser::get_rec FAILED. Record not found.");
		#endif

		return (false);
	}
}

//*******************************************************
//* finds a record by LOOKUP string and current index
//*******************************************************
bool Quser::find_rec (string index_str, bool load_data = true)
{
	#ifdef TRACER_CODE
		DB.Trace ("Quser::find_rec called for index_str: " + index_str);
	#endif

	clear_rec(load_data);			// clear VSAM buffer and all Quser fields

	if (DB.FindRecord (index_str, load_data) == true)
	{
		#ifdef TRACER_CODE
			DB.Trace ("Quser::find_rec OKAY. Now load fields.");
		#endif

		get_all_fields (load_data);
		return (true);
	}
	else
	{
		#ifdef TRACER_CODE
			DB.Trace ("Quser::find_rec FAILED. Record not found.");
		#endif

		return (false);
	}

}

//*******************************************************
//* gets_next_rec in the file
//* by whatever index is currently in use
//*******************************************************
bool Quser::get_next_rec (bool load_data = true)
{
	if (DB.GetNextRecord(load_data) == true)
	{
		#ifdef TRACER_CODE
			DB.Trace ("Quser::get_next_rec OKAY. Now load fields.");
		#endif

		get_all_fields (load_data);
		return (true);
	}
	else
	{
		#ifdef TRACER_CODE
			DB.Trace ("Quser::get_next_rec FAILED. Record not found.");
		#endif

		clear_rec(load_data);
		return (false);
	}
}

//*******************************************************
//* gets_prev_rec in the file
//* by whatever index is currently in use
//*******************************************************
bool Quser::get_prev_rec (bool load_data = true)
{
	if (DB.GetPrevRecord (load_data) == true)
	{
		#ifdef TRACER_CODE
			DB.Trace ("Quser::get_prev_rec OKAY. Now load fields.");
		#endif

		get_all_fields (load_data);
		return (true);
	}
	else
	{
		#ifdef TRACER_CODE
			DB.Trace ("Quser::get_prev_rec FAILED. Record not found.");
		#endif

		clear_rec(load_data);
		return (false);
	}
}


//*******************************************************
//* Logically Clears the Record Buffer *and* FIELDS!!
//*******************************************************
bool Quser::clear_rec (bool load_data = true)
{
	#ifdef TRACER_CODE
		DB.Trace ("Quser::clear_rec called");
	#endif

	DB.FreeVSAMBuffer();			// clear record buffer by clearing VSAM_BUFFER in memory
								// this forces VSAM to create an all new record buffer
								// with empty values for all fields defined in the spec

	if (load_data == true)
		clear_fields ();			// empty all fields (INCLUDING rec_num  PKEY !!)

	return (true);
}

//*******************************************************
//* Deletes a record by PRIMARY KEY
//*******************************************************
bool Quser::delete_rec (string primary_key)
{
	return (DB.DeleteRecord (primary_key));
}

//*******************************************************
//* sets pointer to BOF
//*******************************************************
bool Quser::set_BOF (void)
{
	return (DB.SetBOF());
}

//*******************************************************
//* sets pointer to EOF
//*******************************************************
bool Quser::set_EOF (void)
{
	return (DB.SetEOF());
}

//*******************************************************
//* Get Error Msg
//*******************************************************
string Quser::get_error_msg (void)
{
	return (DB.err_msg);
}


//*******************************************************
//* Get Number of Records in the File
//*******************************************************
long Quser::get_num_recs (void)
{
	return (DB.FileNumRecords ());
}

//*******************************************************
//* returns the number of fields loaded in file spec
//*******************************************************
unsigned int Quser::get_num_fields (void)
{
	return (DB.FileGetNumFields());
}

//*******************************************************
//* sets file to use new INDEX field name
//* use PRIMARY to set to file's p_key index field 0
//*------------------------------------------------------
//* NOTE: AUTO SETS POINTER TO BOF if file open
//*******************************************************
bool Quser::set_index (string new_index_name)
{
	return (DB.SetIndex (new_index_name));
}

//*******************************************************
//* Opens the File
//*******************************************************
bool Quser::open (void)
{
	return (DB.FileOpen ());
}

//*******************************************************
//* Closes the File
//*******************************************************
bool Quser::close (void)
{
	return (DB.FileClose ());
}

//*******************************************************
//* Get Error Flag T/F
//*******************************************************
bool Quser::error (void)
{
	return (DB.error);
}

//*******************************************************
//* Get Error number
//*******************************************************
unsigned int Quser::get_error_num (void)
{
	return (DB.err_num);
}

//*******************************************************
//* FILE Class Destructor - close REQUIRED!
//*******************************************************
Quser::~Quser (void)
{
	#ifdef TRACER_CODE
		DB.Trace ("Quser Class DESTRUCTOR Called.");
	#endif

	DB.FileClose ();	// MUST BE DONE HERE TO PREVENT MEMORY LEAK
}

//*******************************************************
//* Object / Class Copy Constructor 
//*------------------------------------------------------
//*		housekeeping for copied pointers & dynamic arrays
//*------------------------------------------------------
//*		THIS IS ONLY CALLED when class is passed by value
//*     INSTEAD of by reference. If we place the copy
//*     constructor for this class inside the PRIVATE
//*     section of the class def, the class will HAVE to be
//*     passed by reference and not by value which is the
//*     best method! Really dont want to ever have to use
//*     this it can be a very tuff debug when problems
//*     occur. 
//*******************************************************
//Quser::Quser (const ISAM & CopiedObj)
//{
// this will never be called since we have moved the copy 
// constructor for this class to private section of class
// definition. ALWAYS PASS THIS OBJECT BY REFERENCE
// and NOT by value
//
//	trace ("Quser Class COPY CONSTRUCTOR Called.");
//
//
//	trace ("ISAM Class COPY CONSTRUCTOR DONE.");
//
// THIS REPLACES the compiler's default copy object work so everything 
// in the class's constructor must be copied and initialized 
// mnually in here. This is not needed if we pass the object
// by reference and not by value. 
//}

//*******************************************************
//* Get List of VSAM Field Names
//*******************************************************
bool Quser::get_field_name_list (string &list)
{
	long field_count = 0;

	bool result = DB.GetFieldNameList (list, field_count);
	return (result);
}

//*******************************************************
//* Get List of VSAM Field Numbers
//*******************************************************
bool Quser::get_field_num_list (string &list)
{
	long field_count = 0;

	bool result = DB.GetFieldNumList (list, field_count);
	return (result);

}

//*******************************************************
//* Encrypt a Text String
//*******************************************************
string Quser::encrypt (string text, string cypher_key)
{
	return (DB.Encrypt (text, cypher_key));
}

//*******************************************************
//* Decrypt an Encrypted Hex String
//*******************************************************
string Quser::decrypt (string hex_text, string cypher_key)
{
	return (DB.Decrypt (hex_text, cypher_key));
}
